#!/usr/bin/python
# -*- coding: utf-8 -*-

from supy.database import *

###################################################################################################
# Multibase.py
#
# Simple database class for persistent storage of data records in files
# 
# Type of data is defined by a list of attribute definitions
#     Each attribute i has a name attr_names[i] and a type attr_types[i] being either 'string' or 'json'  
#     Type 'json' means that the data can be of any Python type, where
#     'json' data gets converted into a string by using json.dump() when storing in a file
#     Thus, type 'json' is typically used for any data type other than 'string' (e.g., numbers, lists, etc.)
# 
# Each database record consists of a number of data entries, i.e., one data entry for each attribute.
# The first data entry (corresonding to the frist attribute) is defined to be the key of the data record.
# The first attribute is also called the key attribute.
#
# The database stores an arbitrary number of data records in a dictionary (db) 
# Data records can be accessed via their key attribute data value, 
# i.e., get_record(key) which returns a list of all data entries (including the key)
# similarly, get_entry(attr_name,key) returns a single entry, i.e., the data entry corresponding
# to a certain attribute with name attr_name.
# 
# If flag_singleton is set then the database is forced to contain exactly one data record.
#
# 28.3.2013 - 7.4.2013 by Andreas Knoblauch
###################################################################################################


class Multibase_config(Database_config):
    attr_names = ['db_name', 'db'  ]
    attr_types = ['string' , 'json']
    default_record = ['__new_key_', None]
    default_filename = 'dummy.mb'
    db_configs = []  # configuration classes for databases (should all be RAM-only) to be stored in multibase

class Multibase:
    def __init__(self,cfg=Multibase_config, filename=None):
        self.cfg=cfg
        self.filename=filename
        if(self.filename==None): self.filename=self.cfg.default_filename
        self.n_db = len(self.cfg.db_configs)
        self.mb = None
        self.load()

    def isModified(self):      # returns True if at least one database is modified
        res=False
        keys = self.mb.get_list_of_keys()
        for k in keys:
            if self.get_database(k).flagModified>0:
                res=True
        if res:
            self.mb.flagModified=1
        return self.mb.flagModified
            
    def get_database(self,key):
        db=self.mb.get_record_entry("db",key)
        return db

    def load(self):
        try:
            f=open(self.filename,'r')
            f.close()
        except IOError as e:
            # error message is written by call to Database constructor...
            self.mb = Database(self.cfg,self.filename)
            # add empty databases
            print "I will add empty databases to multibase ", self.filename, " ..."
            for i in range(self.n_db):
                db = Database(self.cfg.db_configs[i])
                self.mb.set_record(db.filename,db)
        else:
            self.mb = Database(self.cfg,self.filename)    # load database of databases
            # set the database configs of self.cfg for each of the loaded databases
            # this is necessary because during loading the configs are set to the classes of the stored databases
            # but these classes may differ from the classes in self.cfg!!
            for cfg in self.cfg.db_configs:
                if cfg.default_filename in self.mb.get_list_of_keys():
                    db = self.get_database(cfg.default_filename)
                    db.cfg=cfg

    def save(self, filename=None):
        if(filename!=None):
            self.filename=filename
        self.mb.save(self.filename)
        keys = self.mb.get_list_of_keys()
        for k in keys:
            self.get_database(k).flagModified=0

    def print_multibase(self,indent=0,indent_inc=3):
        list_of_keys = self.mb.get_list_of_keys()
        str_indent = ""
        for i in range(indent)    : str_indent     = str_indent     + " "
        str_indent_inc=str_indent
        for i in range(indent_inc): str_indent_inc = str_indent_inc + " "
        print str_indent+"multibase ", self.filename, " contains the following databases:"
        self.mb.print_database(indent+1,indent_inc)
        keys = self.mb.get_list_of_keys()
        for k in keys:
            db = self.get_database(k)
            db.print_database(indent+1,indent_inc)
            
    def import_data_from_multibase(self,other_mb,db_list=None):
        '''
        imports data from another multibase that may have a (slightly) different configuration file
        (i.e., other databases with other configurations)
        Basically uses Database.import_data_from_database()
        If db_list==None then all databases of the multibase are imported, otherwise only those databases with keys in db_list
        '''
        if db_list==None:
            mb_keys = other_mb.mb.get_list_of_keys()
        else:
            mb_keys = db_list
        my_keys = self.mb.get_list_of_keys()
        for k in mb_keys:                                # scan all keys (database names) of other_mb
            if k in my_keys:                             # if key exists also in self then import data
                my_db=self.get_database(k)
                mb_db=other_mb.get_database(k)
                my_db.import_data_from_database(mb_db)    

    def enableCallbacks(self,flag=1):                   # set callback flags of _all_ databases
        keys = self.mb.get_list_of_keys()
        for k in keys:
            self.get_database(k).enableCallbacks(flag)


#######################################
# Module test     
#######################################
if __name__ == '__main__':

    print "#######################################"
    print "# Test for module supy.multibase.py    "
    print "#######################################"

    class Database_config1(Database_config):
        attr_names     = ['id'    ,'name', 'age'] # list of attribute names
        attr_types     = ['string','string','json'] # list of attribtute types ('string' or 'json')
        default_record = ['__new_key_','Andi',40]                  # for example, for adding a new data record
        default_filename = 'db1'            # default file name of database
        flag_RAM_only    = 1                  # if set then database is kept only in RAM (never read/written to disk) 

    class Database_config2(Database_config):
        attr_names     = ['id'    ,'name', 'address'] # list of attribute names
        attr_types     = ['string','string','string'] # list of attribtute types ('string' or 'json')
        default_record = ['__new_key_','Andi','Froschhausen']   # for example, for adding a new data record
        default_filename = 'db2'            # default file name of database
        flag_RAM_only    = 1                  # if set then database is kept only in RAM (never read/written to disk) 

    class Multibase_config1(Multibase_config):
        default_filename = 'dummy.mb'                           # default file name of database
        db_configs       = [Database_config1, Database_config2] # list of database specification that make up the multibase

    multibase = Multibase(Multibase_config1,"dummy.mb")
    print "\nafter loading..."
    multibase.print_multibase()
    db1 = multibase.get_database("db1")
    db2 = multibase.get_database("db2")
    db1.delete_all_records()
    db2.delete_all_records()
    print "\nafter deleting..."
    multibase.print_multibase()
    db1.set_record("id1","Christian",7)
    db1.set_record("id2","Matthias",5)
    db2.set_record("id1","Andi","Bömenkirch")
    db2.set_record("id2","Kattia","Froschhausen")
    print "\nafter inserting records..."
    multibase.print_multibase()
    multibase.save()




#!/usr/bin/python
# -*- coding: utf-8 -*-

import json
import supy.serialize

###################################################################################################
# database.py
#
# Simple database class for persistent storage of data records in files
# 
# Type of data is defined by a list of attribute definitions
#     Each attribute i has a name attr_names[i] and a type attr_types[i] being either 'string' or 'json'  
#     Type 'json' means that the data can be of any Python type, where
#     'json' data gets converted into a string by using json.dump() when storing in a file
#     Thus, type 'json' is typically used for any data type other than 'string' (e.g., numbers, lists, etc.)
# 
# Each database record consists of a number of data entries, i.e., one data entry for each attribute.
# The first data entry (corresonding to the frist attribute) is defined to be the key of the data record.
# The first attribute is also called the key attribute.
#
# The database stores an arbitrary number of data records in a dictionary (db) 
# Data records can be accessed via their key attribute data value, 
# i.e., get_record(key) which returns a list of all data entries (including the key)
# similarly, get_entry(attr_name,key) returns a single entry, i.e., the data entry corresponding
# to a certain attribute with name attr_name.
# 
# If flag_singleton is set then the database is forced to contain exactly one data record.
#
# 28.3.2013 - 7.4.2013 by Andreas Knoblauch
###################################################################################################

class Database_config:
    attr_names     = ['id'    ,'image' ,'sound1','sound2','info'  ,'imagesize'] # list of attribute names
    attr_types     = ['string','string','string','string','string','json'] # list of attribtute types ('string' or 'json')
    default_record = ['__new_key_','','','','',[100,100]]                  # for example, for adding a new data record
    encoder = supy.serialize.SupyEncoder                                   # json.JSONEncoder for serializing database objects and classes
    decoder = supy.serialize.SupyDecoder                                   # json.JSONDecoder for de-serializing database objects and classes
    readonly_record_keys = []                                              # records with these keys cannot be deleted!!             
    flag_singleton = 0                    # if set then the database contains exactly one data record!!
    default_filename = 'dummy.db'         # default file name of database
    flag_RAM_only    = 0                  # if set then database is kept only in RAM (never read/written to disk) 

class Database:
    def __init__(self,cfg=Database_config, filename=None, db=None, **args):
        self.cfg=cfg
        self.filename=filename
        if(self.filename==None): self.filename=self.cfg.default_filename
        self.flagEnableCallbacks=0                # no callbacks during creation!
        self.attr_n    = len(self.cfg.attr_names) # number of attributes
        self.key_name  = self.cfg.attr_names[0]   # first attribute is key attribute name
        self.db = db
        if(self.db==None):
            self.db = {}                          # empty database
            if not self.cfg.flag_RAM_only:        
                self.load()                       # for non-RAM databased try to load database from file
        self.checkSingleton()                     # check singleton
        self.attr_dict = {self.cfg.attr_names[i]:i for i in range(self.attr_n)}
        self.callbacks_DeletedRecord      =[]     # list of callbacks when a record was deleted
        self.callbacks_InsertedRecord     =[]     # list of callbacks when a record was inserted
        self.callbacks_ModifiedRecord     =[]     # list of callbacks when a record was modified
        self.callbacks_ReorganizedDatabase=[]     # list of callbacks after major reorganization of database 
        self.callbacks_SaveDatabase       =[]     # list of callbacks after call to save (that is intercepted due to flag_RAM_only) 
        self.callbacks_LoadDatabase       =[]     # list of callbacks after call to load (that is intercepted due to flag_RAM_only) 
        self.enableCallbacks()                    # now enable callbacks
        self.flagModified=0                       # if set then there is need to save database

    def enableCallbacks(self,flag=1):             # set callback flag and return old flag
        oldflag=self.flagEnableCallbacks
        self.flagEnableCallbacks=flag
        return oldflag

    def bind(self,eventlist,callback):
        if not isinstance(eventlist,list):
            eventlist = [eventlist]
        for e in eventlist:
            if   e=="<deleted-record>"      : self.callbacks_DeletedRecord      .append(callback)
            elif e=="<inserted-record>"     : self.callbacks_InsertedRecord     .append(callback)
            elif e=="<modified-record>"     : self.callbacks_ModifiedRecord     .append(callback)
            elif e=="<reorganized-database>": self.callbacks_ReorganizedDatabase.append(callback)
            elif e=="<request-to-save-database>": self.callbacks_SaveDatabase   .append(callback)
            elif e=="<request-to-load-database>": self.callbacks_LoadDatabase   .append(callback)
            else:
                raise TypeError, "bind: Unknown event type!"

    def do_callbacks(self,eventlist,key=None):
        if self.flagEnableCallbacks:
            for e in eventlist:
                if   e=="<deleted-record>"   : 
                    for cb in self.callbacks_DeletedRecord      : cb(key)
                elif e=="<inserted-record>"  : 
                    for cb in self.callbacks_InsertedRecord     : cb(key)
                elif e=="<modified-record>"  : 
                    for cb in self.callbacks_ModifiedRecord     : cb(key)
                elif e=="<reorganized-database>": 
                    for cb in self.callbacks_ReorganizedDatabase: cb()
                elif e=="<request-to-save-database>": 
                    for cb in self.callbacks_SaveDatabase       : cb()
                elif e=="<request-to-load-database>": 
                    for cb in self.callbacks_LoadDatabase       : cb()
                else:
                    raise TypeError, "do_callbacks: Unknown event type!"

    def size(self):           
        ''' 
        returns total number of data records that are stored in the database
        '''
        return len(self.db)

    def checkSingleton(self,keep_key=None):                 
        '''
        If flag_singleton is true then enforce the database to be a singleton (with exactly one entry)
        If database is empty a new default record is added
        If database has more then one record then all records are deleted except keep_key or (if None) the first key 
        '''
        if(self.cfg.flag_singleton):
            n=self.size() 
            if(n!=1):
                if n==0:                       # if database is empty then add default entry
                    self.add_default_record()
                elif n>1:                      # if database has more than one entry then delete all except first
                    keys = self.get_list_of_keys()
                    if (keep_key==None) or not(keep_key in keys):
                        for i in range(len(keys)-1):
                            self.delete_record(keys[i+1])
                    else:
                        keys.remove(keep_key)
                        for k in keys:
                            self.delete_record(k)
            
    def get_record(self,key=None):               
        '''
        returns data record with given key; if key=None then return first data record
        '''
        if(key==None):
            key = self.get_first_key() 
        if(key in self.db):
            return [key] + self.db[key]
        else:
            return None

    def get_record_str(self,key=None):               
        '''
        similar as get_record but returns list of strings (obtained by json) 
        '''
        c = self.get_record(key)
        if(c!=None):
            for i in range(self.attr_n):
                if self.cfg.attr_types[i]=='json':
                    c[i]=json.dumps(c[i])
        return c

    def get_record_entry(self,attr_name=None, key=None):
        '''
        returns data record entry with given key and attribute name
        if attr_name and/or key==None then just take first entry and/or key, respectively
        '''
        if(key==None):
            key = self.get_first_key()
        if(attr_name==None):
            attr_name = self.key_name
        res=None
        if(key in self.db) and (attr_name in self.attr_dict):
            idx=self.attr_dict[attr_name]   # attr index
            res=self.get_record(key)[idx]
        return res
            
    def set_record(self,*record):
        '''
        write record to database
        record parameters must be in order of attr_names
        only and final method that can insert or modify a data record in the database 
        thus, this method is responsible for organizing corresponding callbacks
        '''
        assert len(record)==self.attr_n, "Invalid record length in call to set_record() for database " \
            + self.filename + " ! Record length is " + str(len(record)) + ", but attribute list has length " \
            + str(self.attr_n) + "!"
        key=record[0]
        flagInsert=0
        if (key not in self.db) and ((not self.cfg.flag_singleton) or (self.size()==0)):   # is it really an insertion??
            flagInsert=1
        self.db[key]=list(record[1:])
        self.checkSingleton(key)      # may delete a record (if singleton)
        self.flagModified=1
        if(flagInsert>0 ): self.do_callbacks(['<inserted-record>'],key)
        if(flagInsert==0): self.do_callbacks(['<modified-record>'],key) 

    def set_record_str(self,record_str,check_only=0):  # entry_str is list of strings
        '''
        write record given by list of strings to database
        json attributes are transformed via json.loads
        '''
        assert len(record_str)==self.attr_n, "Invalid record length in call set_record_str() for database " + self.filename
        record=[None for i in range(self.attr_n)]
        for i in range(len(record_str)):
            if self.cfg.attr_types[i]=='string':
                record[i]=record_str[i]
            else:
                try:
                    record[i]=json.loads(record_str[i])
                except ValueError as e:
                    print "Error during json.loads with argument ", record_str[i], " !\n"
                    raise e
        if not check_only:
            self.set_record(*record)

    def set_record_entry(self,attr,key,entry):
        idx = self.attr_dict[attr]  
        rec = self.get_record(key)
        rec[idx] = entry
        self.set_record(*rec)

    def set_default_record(self,key=None):
        '''
        Set record with given key to the default record entries of self.cfg.default_record
        '''
        record = [self.cfg.default_record[i] for i in range(self.attr_n)]    # shallow copy
        if(key!=None):
            record[0]=key
        self.set_record(*record)

    def add_default_record(self):
        '''
        adds new record with new unique key to the (otherwise) default record entries of self.cfg.default_record
        and returns new key
        '''
        # (i) find a new dummy-key-name, e.g., "__new_item_XXX__"
        keys = self.get_list_of_keys()
        i=0
        new_key_prefix=self.cfg.default_record[0]
        new_key_postfix="__"
        while 1:
            i=i+1
            new_key = new_key_prefix + str(i) + new_key_postfix
            if not(new_key in keys): 
                break
        # (ii) insert new default entry and return new key
        self.set_default_record(new_key)
        return new_key
        
    def delete_record(self,key):
        '''
        delete record with given key
        returns 1 if delete operation was successfull
        and 0 otherwise 
        '''
        if (key in self.db) and (key not in self.cfg.readonly_record_keys):
            del self.db[key]
            self.do_callbacks(['<deleted-record>'],key) 
            self.checkSingleton()     # may add a new record
            self.flagModified=1
            return 1
        else:
            return 0
        
    def delete_all_records(self):
        '''
        delete all data record of the database
        '''
        old=self.enableCallbacks(0)
        keys = self.get_list_of_keys()    # keep keys
        for k in keys:
            self.delete_record(k)
        self.enableCallbacks(old)
        self.do_callbacks(['<reorganized-database>'])
        self.flagModified=1

    def load(self,filename=None,blockMultipleCallbacks=0):
        '''
        load database from file with name self.filename
        '''
        if self.cfg.flag_RAM_only:
            self.do_callbacks(['<request-to-load-database>'])
        else:
            if filename==None:
                filename=self.filename
            try:
                f=open(filename,'r')
            except IOError as e:
                print "I/O error({0}): {1}".format(e.errno, e.strerror)
                if(self.cfg.flag_singleton):
                    print "I will add single default entry to empty database", self.filename
                else:
                    print "I will set database ", self.filename, " to empty database."
                old=self.enableCallbacks(0)
                self.delete_all_records()       # sets db to {} or adds single default entry
                self.enableCallbacks(old)
                self.flagModified=0
            except:
                print "Unexpected error while loading database ", self.filename, ":", sys.exc_info()[0]
                raise
            else:
                self.db = json.load(f,cls=self.cfg.decoder)
                if(self.cfg.flag_singleton):
                    n=self.size()
                    if(n==0):
                        print "Loaded empty database ", self.filename, "!"
                        print "I will add default data record."
                    if(n>1):
                        print "Singleton Database ", self.filename, " has more than one data record!"
                        print "I will erase all data records except the first one."
                self.checkSingleton()
                self.do_callbacks(['<reorganized-database>'])
                self.flagModified=0
        
    def save(self,filename=None):
        if self.cfg.flag_RAM_only:
            self.do_callbacks(['<request-to-save-database>'])
        else:
            if(filename!=None):
                self.filename=filename
            try:
                f=open(self.filename,'w')
            except IOError as e:
                print "I/O error({0}): {1}".format(e.errno, e.strerror)
                print "Unable to save database ", self.filename
            except:
                print "Unexpected error while trying to save database ", self.filename, ":", sys.exc_info()[0]
                raise
            else:
                json.dump(self.db,f,cls=self.cfg.encoder)
                f.flush()
                self.flagModified=0

    def get_list_of_keys(self):
        keys = self.db.keys()
        keys.sort()
        return keys

    def get_first_key(self):
        keys = self.get_list_of_keys()
        if(len(keys)>0):
            key=keys[0]
        else:
            key=None
        return key

    def print_database(self,indent=0,indent_inc=3):
        list_of_keys = self.get_list_of_keys()
        str_indent = ""
        for i in range(indent)    : str_indent     = str_indent     + " "
        str_indent_inc=str_indent
        for i in range(indent_inc): str_indent_inc = str_indent_inc + " "
        print str_indent+"database ", self.filename, " with attributes ", self.cfg.attr_names, " :" 
        for k in list_of_keys:
            print str_indent_inc, k, " : ", self.db[k] 

    def import_data_from_database(self,db):
        '''
        imports data from another database that may have a (slightly) different configuration file
         - uses only record entries of db that has corresponding attributes in the current database
         - otherwise the default attribute value is used.
         - If a key is already existing then an record from db is not inserted in current database
        Use this method, for example, after you have extended a database configuration by additional attributes,
        and want to use old database files (which were generated with the old configuration)
        '''
        db_attr_names = db.cfg.attr_names    # attribute list of the database db to be imported from
        db_keys = db.get_list_of_keys()      # keys of the database db to be imported from
        my_attr_names = self.cfg.attr_names  # attribute list of self
        my_keys = self.get_list_of_keys()    # keys of self
        # check whether the two attribute lists are identical
        equal_attributes = True
        if (len(db_attr_names)!=len(my_attr_names)):
            equal_attributes = False
        else:
            for i in range(len(db_attr_names)):
                if db_attr_names[i]!=my_attr_names[i]:
                    equal_attributes=False
                    break
        for k in db_keys:
            if not k in my_keys:                            # if key is not yet in self then insert new record...
                if equal_attributes:
                    r = db.get_record(k)                    # if attribute lists are equal then just insert record from db
                    self.set_record(*r)
                else:                                       # otherwise insert only those attribute entries that exist in self:
                    my_r  = self.cfg.default_record             # new record with default values
                    my_r[0] = k                                 # first field is assumed to be the key...
                    self.set_record(*my_r)                      # set new record, modify fields afterwards...
                    for a in db_attr_names:         
                        if a in my_attr_names:                  # if attribute a exists in self then set it in new record
                            db_e = db.get_record_entry(a,k)
                            self.set_record_entry(a,k,db_e)


#######################################
# Module Test 
#######################################

if __name__ == '__main__':
    # create a dummy database
    db = Database(Database_config)
    print "\nAfter load():"
    db.print_database()
    db.delete_all_records()
    print "\nAfter delete():"
    db.print_database()
    db.set_record('item1','../MinotaurusSpiel/trunk/images/default_image.jpg','../MinotaurusSpiel/trunk/sounds/default_sound.wav','../MinotaurusSpiel/trunk/sounds/default_sound.wav','type1',[40,40])
    db.set_record('item2','../MinotaurusSpiel/trunk/images/default_image.jpg','../MinotaurusSpiel/trunk/sounds/default_sound.wav','../MinotaurusSpiel/trunk/sounds/default_sound.wav','type2',[40,40])
    db.set_record('item3','../MinotaurusSpiel/trunk/images/default_image.jpg','../MinotaurusSpiel/trunk/sounds/default_sound.wav','../MinotaurusSpiel/trunk/sounds/default_sound.wav','type3',[40,40])
    db.save()
    print "\nAfter insertion():"
    db.print_database()



#!/usr/bin/python
# -*- coding: utf-8 -*-

###############################################################################
# An extended Frame that makes window menus and tool-bars automatically.
# Use GuiMakerFrameMenu for embedded components (makes frame-based menus).
# Use GuiMakerWindowMenu for top-level windows (makes Tk8.0 window menus).
# See the self-test code (and PyEdit) for an example layout tree format.
###############################################################################
     
import sys
from Tkinter import *                     # widget classes
from types   import *                     # type constants
     
class GuiMaker(Frame):
    menuBar    = []                       # class defaults
    toolBar    = []                       # change per instance in subclasses
    helpButton = 1                        # set these in start() if need self
     
    def __init__(self, parent=None):
        Frame.__init__(self, parent) 
        self.pack(expand=YES, fill=BOTH)        # make frame stretchable
        self.start()                            # for subclass: set menu/toolBar
        self.makeMenuBar()                      # done here: build menu-bar
        self.makeToolBar()                      # done here: build tool-bar
        self.makeWidgets()                      # for subclass: add middle part
     
    def makeMenuBar(self):
        """
        make menu bar at the top (Tk8.0 menus below)
        expand=no, fill=x so same width on resize
        """
        menubar = Frame(self, relief=RAISED, bd=2)
        menubar.pack(side=TOP, fill=X)
     
        for (name, key, items) in self.menuBar:
            mbutton  = Menubutton(menubar, text=name, underline=key)
            mbutton.pack(side=LEFT)
            pulldown = Menu(mbutton)
            self.addMenuItems(pulldown, items)
            mbutton.config(menu=pulldown)
     
        if self.helpButton: 
            Button(menubar, text    = 'Help', 
                            cursor  = 'gumby', 
                            relief  = FLAT, 
                            command = self.help).pack(side=RIGHT)
     
    def addMenuItems(self, menu, items):
        for item in items:                     # scan nested items list
            if item == 'separator':            # string: add separator
                menu.add_separator({})
            elif type(item) == ListType:       # list: disabled item list
                for num in item:
                    menu.entryconfig(num, state=DISABLED)
            elif type(item[2]) != ListType:
                menu.add_command(label     = item[0],         # command: 
                                 underline = item[1],         # add command
                                 command   = item[2])         # cmd=callable
            else:
                pullover = Menu(menu)
                self.addMenuItems(pullover, item[2])          # sublist:
                menu.add_cascade(label     = item[0],         # make submenu
                                 underline = item[1],         # add cascade
                                 menu      = pullover) 
     
    def makeToolBar(self):
        """
        make button bar at bottom, if any
        expand=no, fill=x so same width on resize
        """
        if self.toolBar:
            toolbar = Frame(self, cursor='hand2', relief=SUNKEN, bd=2)
            toolbar.pack(side=BOTTOM, fill=X)
            for (name, action, where) in self.toolBar: 
                Button(toolbar, text=name, command=action).pack(where)
     
    def makeWidgets(self):
        """
        make 'middle' part last, so menu/toolbar 
        is always on top/bottom and clipped last;
        override this default, pack middle any side;
        for grid: grid middle part in a packed frame  
        """
        name = Label(self, 
                     width=40, height=10,
                     relief=SUNKEN, bg='white',   
                     text   = self.__class__.__name__, 
                     cursor = 'crosshair')
        name.pack(expand=YES, fill=BOTH, side=TOP)
     
    def help(self):
        """
        override me in subclass
        """
        from tkMessageBox import showinfo
        showinfo('Help', 'Sorry, no help for ' + self.__class__.__name__)
     
    def start(self): pass  # override me in subclass
     
     
###############################################################################
# For Tk 8.0 main window menubar, instead of a frame
###############################################################################
     
GuiMakerFrameMenu = GuiMaker           # use this for embedded component menus
     
class GuiMakerWindowMenu(GuiMaker):    # use this for top-level window menus
    def makeMenuBar(self):
        menubar = Menu(self.master)
        self.master.config(menu=menubar)
     
        for (name, key, items) in self.menuBar:
            pulldown = Menu(menubar)
            self.addMenuItems(pulldown, items)
            menubar.add_cascade(label=name, underline=key, menu=pulldown)
     
        if self.helpButton: 
            if sys.platform[:3] == 'win':
                menubar.add_command(label='Help', command=self.help)
            else:
                pulldown = Menu(menubar)  # linux needs real pulldown
                pulldown.add_command(label='About', command=self.help)
                menubar.add_cascade(label='Help', menu=pulldown)
     
     
###############################################################################
# Self test when file run stand-alone: 'python guimaker.py'
###############################################################################
     
if __name__ == '__main__':
    from guimixin import GuiMixin            # mixin a help method
     
    menuBar = [ 
        ('File', 0,  
            [('Open',  0, lambda:0),         # lambda:0 is a no-op
             ('Quit',  0, sys.exit)]),       # use sys, no self here
        ('Edit', 0,
            [('Cut',   0, lambda:0),
             ('Paste', 0, lambda:0)]) ]
    toolBar = [('Quit', sys.exit, {'side': LEFT})]
     
    class TestAppFrameMenu(GuiMixin, GuiMakerFrameMenu):
        def start(self):
            self.menuBar = menuBar
            self.toolBar = toolBar
    class TestAppWindowMenu(GuiMixin, GuiMakerWindowMenu):
        def start(self):
            self.menuBar = menuBar
            self.toolBar = toolBar
    class TestAppWindowMenuBasic(GuiMakerWindowMenu):
        def start(self):
            self.menuBar = menuBar
            self.toolBar = toolBar    # guimaker help, not guimixin
     
    root = Tk()
    TestAppFrameMenu(Toplevel())
    TestAppWindowMenu(Toplevel())
    TestAppWindowMenuBasic(root)
    root.mainloop()



#!/usr/bin/python
# -*- coding: utf-8 -*-

from tkinter import *
#import tkmessagebox
#import tkFileDialog
import os,time,datetime
import tkinter.messagebox
import tkinter.filedialog
from PIL import Image, ImageTk, ImageDraw
from supy.utilities import *

##################################################################################################################################################################
##################################################################################################################################################################
# Part I: SupyImageDialog and related classes and procedures:
#         classes: SupyImageDialog, SupyImageDialog_config, SupyImageDialogForm
#         SupyImageDialog: for inputing and clipping a image   
# Data Exchange:
##################################################################################################################################################################
##################################################################################################################################################################
class SupyImageDialog_config:
    text_labels    = ["Label","Filename","Mode","x1" ,"x2" ,"y1" ,"y2" ]
    pos_labels     = [(0,0)  ,(1,0)     ,(2,0)                 ,(3,0)  ,(4,0)  ,(5,0)  ,(6,0)  ]
    entry_width    = [40     ,40        ,None                  ,10     , 10    , 10    , 10    ]
    entry_pos      = [(0,1)  ,(1,1)     ,(2,1)                 ,(3,1)  ,(4,1)  ,(5,1)  ,(6,1)  ]   # format (row,column)
    entry_type     = [['ef'],['ef']     ,['ol',['full','clipped']],['ef'] ,['ef'] ,['ef'] ,['ef'] ]   # 'ef'=entry field; 'ol'=option list
    label_sticky   = 'w'                                                                           # Ausrichtung der Labels (within grid cell)
    entry_sticky   = 'w'                                                                           # Ausrichtung der Entries (within grid cell)
    title          = 'SupyImageDialog'
    button_str     = ['Browse', 'Browse', 'Ok','Apply','Cancel']                                   # Text on buttons 
    button_pos     = [[1,3],[3,3],None,None]
    image_canvas_size = [200, 100]                                                                 # Size of image canvas
    clipframe_mark_size = (5,2)           # side length and line width of marking rectangle                    
    clipframe_mark_color = ('red','red')  # color of square and rectangle of mark symbol

class SupyImageDialogFrame(Frame):
    """
    Default constructor
    """
    def __init__(self, parent, data_str_init=None, cfg=SupyImageDialog_config, paths_relative_to=os.getcwd()):
        # (i) do some basic initialization
        Frame.__init__(self,parent)  
        self.cfg=cfg                                        # configuration data
        self.paths_relative_to = paths_relative_to
        self.data_str=data_str_init                         # string representation of dialog result; format: "Label | Filename | mode | x1 | x2 | y1 | y2" (as could be saved in SQL table)
        self.data_list     = None
        self.data_label    = None                           # label to describe image
        self.data_filename = None                           # filename (inc. path) of image
        self.data_mode     = None                           # either 'full' or 'clipped': in former case the raw image is used, in the latter case a clip (see below) 
        self.data_x1       = None                           # (x1,x2,y1,y2) define a rectangle to clip from the original image
        self.data_x2       = None
        self.data_y1       = None
        self.data_y2       = None
        # (ii) allocate input widgets
        n_ip = len(self.cfg.text_labels)                     # number of input widgets
        self.leftframe = Frame(self, bd=2, relief=GROOVE)
        self.leftframe.pack(side=LEFT)
        self.rightframe = Frame(self)
        self.rightframe.pack(side=RIGHT,expand=YES, fill=BOTH)
        Button(self.leftframe,text=self.cfg.button_str[0],command=self.onBrowseFilename).grid(row=self.cfg.button_pos[0][0],column=self.cfg.button_pos[0][1])
        Button(self.leftframe,text=self.cfg.button_str[1],command=self.onBrowseClip    ).grid(row=self.cfg.button_pos[1][0],column=self.cfg.button_pos[1][1])        
        for i in range(n_ip): Label(self.leftframe,text=self.cfg.text_labels[i],anchor=self.cfg.label_sticky).grid(row=self.cfg.pos_labels[i][0],column=self.cfg.pos_labels[i][1],sticky=self.cfg.label_sticky)  # labels
        self.entry_vars = [StringVar() for i in range(n_ip)] # string-type entry variables 
        self.entry_field = [None for i in range(n_ip)]
        for i in range(n_ip):
            if   self.cfg.entry_type[i][0]=='ef':      # entry-field?
                entry = Entry(self.leftframe, textvariable=self.entry_vars[i], width=self.cfg.entry_width[i], bg="white", disabledforeground='black')
            elif self.cfg.entry_type[i][0]=='ol':      # option-list?
                entry = OptionMenu(self.leftframe,self.entry_vars[i],*self.cfg.entry_type[i][1])
            else:
                assert 0, "Wrong entry type: self.cfg.entry_type[i][0]="+str(self.cfg.entry_type[i][0])+" for i="+str(i)
            entry.grid(row=self.cfg.entry_pos[i][0],column=self.cfg.entry_pos[i][1],sticky=self.cfg.entry_sticky)
            self.entry_field[i]=entry
        self.dictEntryVar = {'label':self.entry_vars[0], 'fname':self.entry_vars[1], 'mode':self.entry_vars[2], \
                             'x1':self.entry_vars[3], 'x2':self.entry_vars[4], 'y1':self.entry_vars[5], 'y2':self.entry_vars[6]}
        # (iii) allocate image widgets (on rightframe)
        self.imcanvas = Canvas(self.rightframe,relief=SUNKEN)
        self.imcanvas.config(width=self.cfg.image_canvas_size[0], height=self.cfg.image_canvas_size[1])
        self.imcanvas.config(scrollregion=(0,0,self.cfg.image_canvas_size[0],self.cfg.image_canvas_size[1]))
        self.sbarx, self.sbary = Scrollbar(self.rightframe,orient=HORIZONTAL), Scrollbar(self.rightframe,orient=VERTICAL)
        self.sbarx.config(command=self.imcanvas.xview)
        self.sbary.config(command=self.imcanvas.yview)
        self.imcanvas.config(xscrollcommand=self.sbarx.set,yscrollcommand=self.sbary.set)
        self.sbarx.pack(side=BOTTOM, fill=X)
        self.sbary.pack(side=RIGHT, fill=Y)
        self.imcanvas.pack(side=RIGHT,expand=YES,fill=BOTH)
        self.imcanvas.bind("<Button-1>",self.onPressedB1)
        self.imcanvas.bind("<B1-Motion>",self.onMovedB1)
        self.imcanvas.bind("<ButtonRelease-1>",self.onReleasedB1)
        self.phim, self.draw = None, None
        self.imcanvas_state='IDLE'
        # (iv) set data 
        self.setData_from_str(self.data_str)           # set data fields from self.data_str
        if self.phim: self.sbarCenterView()            # center view on clip (i.e. adjust srollbars accordingly)

    def setData_from_str(self,data_str):                         # split data_str in components and save in object fields
        if data_str:
            s=data_str.split('|')
            assert len(s)==7, "Assertion error in SupyImageDialog.setData_from_str(data_str): Parameter data_str='" + data_str + "' must have format 'Label | Filename | mode | x1 | x2 | y1 | y2' !"
            self.data_label   =str(s[0])
            self.data_filename=str(s[1])
            self.data_mode    =str(s[2])
            self.data_x1      =int(s[3])
            self.data_x2      =int(s[4])
            self.data_y1      =int(s[5])
            self.data_y2      =int(s[6])
            self.data_str = data_str
            self.data_list = [self.data_label, self.data_filename, self.data_mode, self.data_x1, self.data_x2, self.data_y1, self.data_y2]
        else:
            self.data_label, self.data_filename, self.data_mode = "", "",str(self.cfg.entry_type[2][0])
            self.data_x1, self.data_x2, self.data_y1, self.data_y2 = 0,0,0,0
        self.dictEntryVar['label'].set(self.data_label)
        self.dictEntryVar['fname'].set(self.data_filename)
        self.dictEntryVar['mode' ].set(self.data_mode)
        self.dictEntryVar['x1'   ].set(str(self.data_x1))
        self.dictEntryVar['x2'   ].set(str(self.data_x2))
        self.dictEntryVar['y1'   ].set(str(self.data_y1))
        self.dictEntryVar['y2'   ].set(str(self.data_y2))
        self.setImage()

    def setData_from_form(self):                               # evaluate form inputs and store them in result data fields
        self.data_label    = str(self.dictEntryVar['label'].get())
        self.data_filename = str(self.dictEntryVar['fname'].get())
        self.data_mode     = str(self.dictEntryVar['mode' ].get())
        self.data_x1       = int(self.dictEntryVar['x1'   ].get())
        self.data_x2       = int(self.dictEntryVar['x2'   ].get())
        self.data_y1       = int(self.dictEntryVar['y1'   ].get())
        self.data_y2       = int(self.dictEntryVar['y2'   ].get())
        self.data_str      = self.data_label + '|' + self.data_filename + '|' + self.data_mode + '|' + str(self.data_x1) + '|' + str(self.data_x2) + '|' + str(self.data_y1) + '|' + str(self.data_y2) 
        self.data_list = [self.data_label, self.data_filename, self.data_mode, self.data_x1, self.data_x2, self.data_y1, self.data_y2]

    def load_image(self,fname):   # load image with filename fname and set to self.phim
        self.phim,self.draw=None,None
        if(fname):
            try:
                self.image = Image.open(search_path(fname,self.paths_relative_to))
            except IOError as e:
                msg="I/O error({0}): {1}".format(e.errno, e.strerror) + "\nsupy.forms.SupyImageDialogFrame.load_image: Cannot load image " + fname + " !"
                tkinter.messagebox.showerror(title="Submit Error", parent=self, message=msg)
            else:
                self.draw = ImageDraw.Draw(self.image)
                self.phim = ImageTk.PhotoImage(self.image)
        return self.phim,self.draw

    def setImage(self):
        # (i) load image
        self.load_image(self.data_filename)
        # (ii) draw image on canvas
        if(not self.imcanvas is None):
            self.imcanvas.delete('all');
            szx,szy=0,0
            if(not self.phim is None):
                szx,szy=self.phim.width(), self.phim.height()
                self.imcanvas.config(scrollregion=(0,0,szx,szy))
                self.imcanvas.create_image(0,0,image=self.phim,anchor=NW)
                if(self.dictEntryVar['mode'].get()=='clipped'):
                    self.imcanvas.create_rectangle(self.data_x1,self.data_y1,self.data_x2,self.data_y2,width=self.cfg.clipframe_mark_size[1],outline=self.cfg.clipframe_mark_color[1])
                    dxy=int(self.cfg.clipframe_mark_size[0]*0.5) 
                    self.imcanvas.create_rectangle(self.data_x1-dxy,self.data_y1-dxy,self.data_x1+dxy,self.data_y1+dxy,width=self.cfg.clipframe_mark_size[1],outline=self.cfg.clipframe_mark_color[0])
                    self.imcanvas.create_rectangle(self.data_x2-dxy,self.data_y2-dxy,self.data_x2+dxy,self.data_y2+dxy,width=self.cfg.clipframe_mark_size[1],outline=self.cfg.clipframe_mark_color[0])

    def sbarCenterView(self):     # center scrollbar position on clipped image (if necessary)
        x1,x2,y1,y2 = self.data_x1, self.data_x2, self.data_y1, self.data_y2     # bounding box clipped image
        dx,dy       = abs(x2-x1), abs(y2-y1)                                     # width, height of clipped image
        imx,imy     = self.phim.width(), self.phim.height()                      # width, height of original image
        cvx,cvy     = self.imcanvas.winfo_width(), self.imcanvas.winfo_height()  # canvas size
        if(cvx<2) : cvx=self.imcanvas.winfo_reqwidth()
        if(cvy<2) : cvy=self.imcanvas.winfo_reqheight()
        fx,fy,padx,pady=0,0,0,0   
        if(x2>cvx)and(imx>cvx):
            padx=int(0.5*(cvx-dx))
            if(padx<0):
                padx=0
            fx=(x1-padx)/imx
        if(y2>cvy)and(imy>cvy):
            pady=int(0.5*(cvy-dy))
            if(pady<0):
                pady=0
            fy=(y1-pady)/imy
        self.imcanvas.xview_moveto(fx)
        self.imcanvas.yview_moveto(fy)

    def onBrowseFilename(self):
        s=self.dictEntryVar['fname'].get()   
        filename  = os.path.basename(s)
        directory = os.path.dirname(s)
        if directory=="":
            directory="."
        s=tkinter.filedialog.askopenfilename(initialdir=directory, initialfile=filename)
        if(not s is None)and(s!="")and(s):
            s=os.path.relpath(s,self.paths_relative_to)
            self.dictEntryVar['fname'].set(s)
            self.data_filename=s
            self.setImage()

    def onPressedB1(self,event):    # pressed left button on image canvas
        # check if clicked on lower-right (x2,y2)-mark
        szx,szy=self.phim.width(), self.phim.height()
        x,y=int(event.x+szx*self.sbarx.get()[0]),int(event.y+szy*self.sbary.get()[0])
        tol=2*self.cfg.clipframe_mark_size[0]
        if (abs(x-self.data_x1)<tol) and (abs(y-self.data_y1)<tol):
            # click on upper-left mark: then move upper-left mark
            self.imcanvas_state='MOVE_UL'
            self.onMovedB1(event)
        elif (abs(x-self.data_x2)<tol) and (abs(y-self.data_y2)<tol):
            # click on lower-right mark: then move lower-right mark
            self.imcanvas_state='MOVE_LR'
            self.onMovedB1(event)
        else:
            # else: set upper-left and then move lower-right mark
            self.dictEntryVar['x1'].set(str(x))
            self.dictEntryVar['y1'].set(str(y))
            self.data_x1,self.data_y1=x,y
            self.imcanvas_state='MOVE_LR'
            self.onMovedB1(event)

    def onMovedB1(self,event):      # moved left button on image canvas
        if(self.imcanvas_state!='IDLE'):
            szx,szy=self.phim.width(), self.phim.height()
            x,y=int(event.x+szx*self.sbarx.get()[0]),int(event.y+szy*self.sbary.get()[0])
            if(self.imcanvas_state=='MOVE_LR'):
                # move lower-right mark
                self.dictEntryVar['x2'].set(str(x))
                self.dictEntryVar['y2'].set(str(y))
                self.data_x2,self.data_y2=x,y
            else: 
                # move upper-left mark
                self.dictEntryVar['x1'].set(str(x))
                self.dictEntryVar['y1'].set(str(y))
                self.data_x1,self.data_y1=x,y
            self.setImage()

    def onReleasedB1(self,event):   # released left button on image canvas
        self.onMovedB1(event)
        self.imcanvas_state='IDLE'

    def onBrowseClip(self):
        pass

    def onApply(self):
        self.setData_from_form()
        self.setImage()


class SupyImageDialog:
    def __init__(self, parent, data_str_init=None, cfg=SupyImageDialog_config):           # for parameters see SupyListboxForm_config
        # basic initialization
        self.cfg = cfg
        self.data_list=None
        self.data_str =None
        self.box = Frame(parent)                             # container frame
        self.box.pack(expand=YES, fill=X)
        # create ImageDialogFrame
        self.imdia = SupyImageDialogFrame(self.box,data_str_init,cfg) 
        self.imdia.pack(side=TOP) #, expand=Y, fill=X)      
        # create control button frame
        self.buttonframe = Frame(self.box)
        Button(self.buttonframe,text=self.cfg.button_str[4],command=self.onCancel).pack(side=RIGHT)
        Button(self.buttonframe,text=self.cfg.button_str[2],command=self.onSubmit).pack(side=RIGHT)
        Button(self.buttonframe,text=self.cfg.button_str[3],command=self.onApply).pack(side=LEFT)
        self.box.master.bind('<Return>', (lambda event: self.onSubmit()))
        self.box.master.bind('<Shift-Return>', (lambda event: self.onSubmit()))
        self.buttonframe.pack(side=TOP)

    def onSubmit(self):                                      # override this if necessary
        self.imdia.setData_from_form()
        self.data_list = self.imdia.data_list 
        self.data_str  = self.imdia.data_str 
        self.box.destroy()
     
    def onApply(self):
        self.imdia.onApply()

    def onCancel(self):                                      # override if needed
        self.data_list=None
        self.data_str=None
        self.box.destroy()
 
def askSupyImageDialog(cfg=SupyImageDialog_config, parent=None, title=None, data_str_init=None):
    if title is None: title=cfg.title
    top=Toplevel(parent)
    top.grab_set()                         # make window modal (i.e., redirect all inputs from parent window to listbox dialog)
    if parent: top.transient(parent)       # make window transient, e.g., minimize with parent etc.
    if title: top.title(title) 
    f=SupyImageDialog(top,data_str_init,cfg)
    top.wait_window(f.box)
    top.destroy()
    return f.data_str, f.data_list         # return results
    

##################################################################################################################################################################
##################################################################################################################################################################
# Part II: SupyTextFrame/Form: A Tkinter Text Widget with scrollbars
##################################################################################################################################################################
##################################################################################################################################################################

class SupyTextFrame_cfg:
    readonly=False
    height=24
    width=80
    wrap=NONE        # either NONE or CHAR
    catch_focus=True  # catch Focus to textfield after construction??

class SupyTextFrame(Frame):
    def __init__(self,parent=None,text_str='',cfg=SupyTextFrame_cfg, readonly=None, height=None, width=None, wrap=None, catch_focus=None, textvariable=None):
        Frame.__init__(self,parent)
        self.cfg=cfg
        if(readonly    is None): readonly   =cfg.readonly
        if(height      is None): height     =cfg.height
        if(width       is None): width      =cfg.width
        if(wrap        is None): wrap       =cfg.wrap
        if(catch_focus is None): catch_focus=cfg.catch_focus
        self.readonly   =readonly
        self.height     =height
        self.width      =width
        self.wrap       =wrap
        self.catch_focus=catch_focus
        self.textvariable = textvariable
        self.text=Text(self,wrap=self.wrap,relief=SUNKEN,width=self.width,height=self.height)
        self.sbarx, self.sbary = Scrollbar(self,orient=HORIZONTAL), Scrollbar(self,orient=VERTICAL)
        self.sbarx.config(command=self.text.xview)
        self.sbary.config(command=self.text.yview)
        self.text.config(xscrollcommand=self.sbarx.set,yscrollcommand=self.sbary.set)
        self.sbarx.pack(side=BOTTOM, fill=X)
        self.sbary.pack(side=RIGHT, fill=Y)
        self.text.pack(side=LEFT, expand=YES, fill=BOTH)
        if (text_str=='' or text_str is None)and (not textvariable is None and textvariable.get()!=''): 
            self.settext(textvariable.get())
        else:
            self.settext(text_str)

    def configure(self,state=None,textvariable=None):
        if(not state is None):
            self.text.configure(state=state)
        if(not textvariable is None):
            self.textvariable=textvariable
            self.settext(textvariable.get(),set_textvariable=False)

    def settext(self,text_str='',set_textvariable=True):        # set text given by text_str
        oldstate = self.text.cget('state')     # save old state
        self.text.configure(state=NORMAL)      # set state NORMAL to be able to reset text
        if(self.textvariable and set_textvariable): self.textvariable.set(text_str)
        self.text.delete('1.0',END)
        self.text.insert('1.0',text_str)
        self.text.mark_set(INSERT, END)
        if(self.readonly):
            self.text.configure(state=DISABLED)
        else:
            self.text.mark_set(INSERT, END)
            if self.catch_focus: self.text.focus()
            self.text.configure(state=oldstate)

    def gettext(self):                    # return text of text-widget
        if(self.textvariable): self.textvariable.set(self.text.get('1.0', END+'-1c'))
        return self.text.get('1.0', END+'-1c')    # first through last
        

class SupyTextForm_cfg(SupyTextFrame_cfg):
    txt_button_submit="Submit"
    txt_button_cancel="Cancel"
    title="SupyTextForm"

class SupyTextForm:
    """
    Input form using a Text widget
    Parameters:
       parent = master widget
       default_text = initial text
    """
    def __init__(self, parent, default_text=None, cfg=SupyTextFrame_cfg, readonly=None, height=None, width=None, wrap=None):    
        self.cfg=cfg
        self.box = Frame(parent)
        self.box.pack(expand=YES, fill=X)
        self.textbox=SupyTextFrame(self.box,default_text,cfg,readonly,height,width,wrap)
        self.textbox.pack(side=TOP,expand=YES,fill=X)
        Button(self.box, text=self.cfg.txt_button_cancel, command=self.onCancel).pack(side=RIGHT)
        Button(self.box, text=self.cfg.txt_button_submit, command=self.onSubmit).pack(side=RIGHT)
        self.result=None

    def onSubmit(self):
        self.result=self.textbox.gettext()
        self.box.destroy()
     
    def onCancel(self):                                       # override if need
        self.results=None
        self.box.destroy()
    
def askSupyTextForm(default_text=None, cfg=SupyTextForm_cfg, parent=None, title=None, readonly=None, height=None, width=None, wrap=None, grab_set=True):      
    if title is None: title=cfg.title
    top=Toplevel(parent)
    if grab_set: top.grab_set()            # make window modal (i.e., redirect all inputs from parent window to listbox dialog)
    if parent: top.transient(parent)       # make window transient, e.g., minmize with parent etc.
    if title: top.title(title) 
    f=SupyTextForm(top,default_text,cfg,readonly,height,width,wrap)
    top.wait_window(f.box)
    top.destroy()
    return f.result


##################################################################################################################################################################
##################################################################################################################################################################
# Part III: SupyForm: a reusable form class 
#           python3-version of moudule supy.forms.py
##################################################################################################################################################################
##################################################################################################################################################################

class SupyForm:                                           # add non-modal form box
    def __init__(self, labels, default_entries=None, parent=None,labelanchor='w',labelwidth=None,entrysize=40,pos=None,
                 flagButtons=1,options=None,paths_relative_to=os.getcwd(),browse_callback=None,
                 help_event="<F1>",help_texts=None,help_callback=None):        # pass field labels list
        '''
        labels: list of strings where labels[i] is the label of the i-th input field
        default_entries: list of default entries that are displayed initially in the input field
        parent: parent widget
        labelanchor: alignment of label
        labelwidth: width of label fields, either a None, a single number, or a list (for each input field)
        entrysize: width of entry fields, either a single number or a list (for each input field)
        pos: either None or list; pos[i]=(x,y) is grid position for i-th label etc.  
        flagButtons: If >0 then buttons are displayed for submit/cancel actions 
        options: either None or a list where options[i] is either 'filename' or 'filename-saveas' or 'filename-dir' or None or a string list for a Optionmenu
        paths_relative_to: save browsed paths relative to this path
        browse_callback(idx): function that is called after selecting a filename, parameter idx is field index
        help_event: event that evokes help function (e.g., pressing F1 over the label)
        help_texts: either None or list of help strings for each label
        help_call : either None (then default call is used) or function 
        '''
        self.parent=parent
        if(pos is None):
            pos = [(0,y) for y in range(len(labels))]
        if(not labelwidth):
            labelwidth=max([len(s) for s in labels])
        if not isinstance(labelwidth,list):
            labelwidth = [labelwidth for y in range(len(labels))]
        if not isinstance(entrysize,list):
            entrysize = [entrysize for y in range(len(labels))]
        self.options = options
        self.paths_relative_to = paths_relative_to
        self.browse_callback=browse_callback
        self.browsebuttons = []
        self.labels = labels
        box = Frame(parent)
        box.pack(expand=YES, fill=X)
        formbox = Frame(box, bd=2, relief=GROOVE)     # formbox will have grid layout
        self.formbox=formbox
        formbox.pack(side=TOP, expand=Y, fill=X)      # runs onSubmit method
        self.content = {}                 # dictionary of input entries
        self.results = None               # final results as set by onSubmit()
        self.entry_vars = [StringVar() for l in labels]
        for sv in self.entry_vars: sv.set("")
        for i in range(len(labels)):
            row_label,column_label=pos[i][1], pos[i][0]
            lb=Label(formbox, text=labels[i], anchor=labelanchor, width=labelwidth[i])
            lb.grid(row=row_label,column=column_label)
            if(not help_texts is None) and (not help_texts[i] is None):
                if(help_callback is None):
                    help_callback=self.displayHelpMessage_default
                lb.bind(help_event,lambda ev,txt=help_texts[i]: help_callback(txt))
            if(options is None)or(options[i] is None)or(options[i]=='filename')or(options[i]=='filename-saveas')or(options[i]=='filename-dir')or(options[i]==[]):                
                entry = Entry(formbox, textvariable=self.entry_vars[i], width=entrysize[i], 
                              bg="white", disabledforeground='black')
                self.content[labels[i]] = entry
            else:
                entry = OptionMenu(formbox,self.entry_vars[i],*options[i])
                entry.config(width=int(0.88*entrysize[i]), disabledforeground='black')
                self.content[labels[i]] = self.entry_vars[i]
            if default_entries and (len(default_entries)>=(i+1)):
                self.setEntryText(i,default_entries[i])                
            entry.grid(row=row_label,column=column_label+1) # side=TOP, expand=YES, fill=X)
            if(not options is None)and(options[i] in ['filename','filename-saveas','filename-dir']):
                b=Button(formbox,text='Browse',command=(lambda i=i,label=labels[i]:self.onBrowse(i,label)))
                b.grid(row=row_label,column=column_label+2)
                self.browsebuttons.append(b)   
        if(flagButtons>0):
            Button(box, text='Cancel', command=self.onCancel).pack(side=RIGHT)
            Button(box, text='Submit', command=self.onSubmit).pack(side=RIGHT)
            box.master.bind('<Return>', (lambda event: self.onSubmit()))
        self.box=box
        self.formbox=formbox
        self.labels=labels

    def displayHelpMessage_default(self,text):
        w=Toplevel(self.parent)
        if self.parent: w.transient(self.parent)     # make window transient, e.g., minmize with parent etc.
        w.title("Help")
        l=Label(w,text=text)
        l.pack()
        b=Button(w,text="Close",command=w.destroy)
        b.pack()
        #tkMessageBox.showinfo("Help", text, parent=w)

    def setEntryText(self,idx,txt):
        if isinstance(self.content[self.labels[idx]],Entry):
            self.entry_vars[idx].set(txt)
            self.content[self.labels[idx]].configure(textvariable=self.entry_vars[idx])
        else:
            if (txt in self.options[idx]):
                self.entry_vars[idx].set(txt)

    def disableWidgets(self):
        for w in self.content.values():
            w.configure(state=DISABLED)
        for b in self.browsebuttons:
            b.configure(state=DISABLED)
        
    def enableWidgets(self):
        for w in self.content.values():
            w.configure(state=NORMAL)
        for b in self.browsebuttons:
            b.configure(state=NORMAL)
        
    def onBrowse(self,i,label):
        s=self.content[label].get()   
        filename  = os.path.basename(s)
        directory = os.path.dirname(s)
        if directory=="":
            directory="."
        if self.options[i]=='filename':
            s=filedialog.askopenfilename(initialdir=directory, initialfile=filename)
        elif self.options[i]=='filename-saveas':
            s=filedialog.asksaveasfilename(initialdir=directory, initialfile=filename)
        elif self.options[i]=='filename-dir':
            s=filedialog.askdirectory(initialdir=directory)
        #print("raw s=",s)
        if(not s is None)and(s!="")and(s):
            #if self.options[i]=='filename-dir':
            #s_rel=os.path.relpath(s,self.paths_relative_to)
            #s_base=os.path.basename(s)
            #print("s_rel=",s_rel, "s_base=",s_base)
            #if self.options[i]=='filename-dir':
            #s=os.path.relpath(s,self.paths_relative_to)
            #else:
            s=os.path.relpath(s,self.paths_relative_to)
            #print("s=",s)
            self.setEntryText(i,s)
            if(not self.browse_callback is None):
                self.browse_callback(i)

    def onSubmit(self):                                      # override this
        #for key in self.content.keys():                      # user inputs in 
        #    print(key, '\t=>\t', self.content[key].get())    # self.content[k]
        self.results = [self.content[l].get() for l in self.labels] 
        self.box.destroy()
     
    def onCancel(self):                                      # override if need
        self.results=None
        self.box.destroy()
     
#class DynamicPyFeForm(PyFeForm):
#    def __init__(self, labels=None):
#        labels = raw_input('Enter field names: ').split()
#        Form.__init__(self, labels)
#    def onSubmit(self):
#        print 'Field values...'
#        Form.onSubmit(self)           
#        self.onCancel()
        
def askSupyForm(labels,default_entries=None,parent=None,title=None,labelanchor='w',labelwidth=None,entrysize=40,pos=None,
                flagButtons=1,options=None,paths_relative_to=os.getcwd(),browse_callback=None,
                help_event="<F1>",help_texts=None,help_callback=None):      
    top=Toplevel(parent)
    if parent: top.transient(parent)       # make window transient, e.g., minmize with parent etc.
    if title: top.title(title) 
    f=SupyForm(labels,default_entries,top,labelanchor,labelwidth,entrysize,pos,
               flagButtons,options,paths_relative_to,browse_callback,
               help_event,help_texts,help_callback)
    top.wait_window(f.box)
    top.destroy()
    return f.results


##################################################################################################################################################################
##################################################################################################################################################################
# Part IV: DateSlider: slider for dates
#          python3-version of moudule supy.forms.py
##################################################################################################################################################################
##################################################################################################################################################################

class DateSlider(Frame):
    def __init__(self, parent, from_date=datetime.date(2012,1,1),to_date=datetime.date(2012,12,31),init_date=datetime.date(2012,1,1),orient_arg='horizontal',command=None,*args):
        Frame.__init__(self,parent)
        dd=to_date-from_date
        self.var=IntVar()
        self.var.set((init_date-from_date).days)
        self.from_date=from_date
        self.current_date=init_date
        self.to_date=to_date
        self.command=command
        self.slider = Scale(self,from_=0,to_=dd.days,command=self.onMove,variable=self.var,*args,orient=orient_arg)
        self.slider.pack(side=RIGHT)
        self.label = Label(self,text="Halleluja")
        self.label.pack(side=LEFT)
        self.onMove(self.var.get(),1)
        
    def onMove(self,value,initFlag=0):
        self.current_date=self.from_date+datetime.timedelta(self.var.get())
        self.str_date = "%02d/%02d/%04d" % (self.current_date.month, self.current_date.day, self.current_date.year)
        self.label.configure(text=self.str_date)
        if (not self.command is None) and (initFlag==0):
            self.command(value)

    def get(self):
        return self.current_date


    
##################################################################################################################################################################
##################################################################################################################################################################
# Part V: SupySliderForm: Input form using Sliders
#         python3-version of moudule supy.forms.py
##################################################################################################################################################################
##################################################################################################################################################################

class SupySliderForm:
    """
    Input form using Sliders
    Parameters:
       parent = master widget
       labels = list of name lables for sliders
       default_values = initial slider values (either float or date)
       ranges = list of triples (v_min,v_max,delta_v)
    """
    def __init__(self, parent, labels, default_values, ranges=None, labelanchor='w',command=None,labelwidth=None,sliderwidth=40):        # pass field labels list
        if(not labelwidth):
            labelwidth=max([len(s) for s in labels])
        box = Frame(parent)
        box.pack(expand=YES, fill=X)
        sliderbox=Frame(box)
        sliderbox.pack(side=TOP,expand=YES,fill=X)

        #rows = Frame(box, bd=2, relief=GROOVE)        # box has rows, button
        #lcol = Frame(rows)                            # rows has lcol, rcol
        #rcol = Frame(rows)                            # button or return key,
        #rows.pack(side=TOP, expand=Y, fill=X)         # runs onSubmit method
        #lcol.pack(side=LEFT)
        #rcol.pack(side=RIGHT, expand=Y, fill=X)
        self.labels=labels
        self.content = {}                 # dictionary of sliders
        self.results = None               # final results as set be onSubmit()
        for i in range(len(labels)):
            label = labels[i]
            Label(sliderbox, text=label+": ", anchor=labelanchor, width=labelwidth).grid(row=i,column=0)
            default_value=default_values[i]
            range_=ranges[i]
            if type(default_value)==datetime.date:
                slider=DateSlider(sliderbox,range_[0],range_[1],default_value,command=self.onUpdate)
            else:
                slider=Scale(sliderbox,from_=range_[0],to_=range_[1],resolution=range_[2],command=self.onUpdate,orient='horizontal')
                slider.set(default_value)
            slider.grid(row=i,column=1) #side=TOP, expand=YES, fill=X)
            self.content[label] = slider
        Button(box, text='Cancel', command=self.onCancel).pack(side=RIGHT)
        Button(box, text='Submit', command=self.onSubmit).pack(side=RIGHT)
        box.master.bind('<Return>', (lambda event: self.onSubmit()))
        self.box=box
        self.sliderbox=sliderbox
        self.command=command
   
    def onUpdate(self,value=0):                                   # override this
        self.results = [self.content[l].get() for l in self.labels] 
        if(self.command):
            self.command()
     
    def onSubmit(self):
        for key in self.content.keys():                      # user inputs in 
            print(key, '\t=>\t', self.content[key].get())    # self.content[k]
        self.onUpdate()
        self.box.destroy()
     
    def onCancel(self):                                       # override if need
        self.results=None
        self.box.destroy()
    
def askSupySliderForm(labels,default_values,ranges,parent=None,title=None,labelanchor='w',command=None,labelwidth=None,sliderwidth=40):      
    top=Toplevel(parent)
    if parent: top.transient(parent)       # make window transient, e.g., minmize with parent etc.
    if title: top.title(title) 
    f=SupySliderForm(top,labels,default_values,ranges,labelanchor,command,labelwidth,sliderwidth)
    top.wait_window(f.box)
    top.destroy()
    return f.results

##################################################################################################################################################################
##################################################################################################################################################################
# Part VI: Module test
##################################################################################################################################################################
##################################################################################################################################################################
if __name__ == '__main__':
    ##################################################################################################################################################################
    print("\nModule test of supylib module supy.forms3.py:")
    print("------------------------------------------------")
    print("\nPart (i): Test of SupyImageDialog:")
    print("----------------------------------")
    ##################################################################################################################################################################
    root=Tk()
    class mySupyImageDialog_config(SupyImageDialog_config):
        text_labels    = ["Label","Filename","Mode","x1" ,"x2" ,"y1" ,"y2" ]
        pos_lables     = [(0,0)  ,(1,0)     ,(2,0)                 ,(3,0)  ,(4,0)  ,(5,0)  ,(6,0)  ]
        entry_width    = [40     ,40        ,None                  ,10     , 10    , 10    , 10    ]
        entry_pos      = [(0,1)  ,(1,1)     ,(2,1)                 ,(3,1)  ,(4,1)  ,(5,1)  ,(6,1)  ]   # format (row,column)
        entry_type     = [['ef'],['ef']     ,['ol',['full','clipped']],['ef'] ,['ef'] ,['ef'] ,['ef'] ]   # 'ef'=entry field; 'ol'=option list
        label_sticky   = 'w'                                                                           # Ausrichtung der Labels (within grid cell)
        entry_sticky   = 'w'                                                                           # Ausrichtung der Entries (within grid cell)
        title          = 'SupyImageDialog'
        button_str     = ['Browse', 'Browse', 'Ok','Apply','Cancel']                                   # Text on buttons 
        button_pos     = [[1,3],[3,3],None,None]
        image_canvas_size = [500, 300]                                                                 # Size of image canvas
        clipframe_mark_size = (5,2)           # side length and line width of marking rectangle                    
        clipframe_mark_color = ('red','red')  # color of square and rectangle of mark symbol

    res_str, res_list = askSupyImageDialog(mySupyImageDialog_config,parent=root,title="Mein Titel...",data_str_init="Studenten WS 13/14|bild1.jpg|clipped|1180|1300|820|846")
    print("Result: res_str=", res_str, " res_list=", res_list)

    ##################################################################################################################################################################
    print("\nPart (ii): Test of SupyTextForm")
    print("--------------------------------")
    ##################################################################################################################################################################
    def_text='Hallo Zeile 1\nDies ist die zweite Zeile\n...und dies die dritte und letzte Zeile!!!!!!!'
    res_str=askSupyTextForm(def_text,wrap=NONE)
    print("Result: res_str=", res_str)

    ##################################################################################################################################################################
    print("\nPart (iii): Test of SupyForm")
    print("--------------------------------")
    ##################################################################################################################################################################
    s=DateSlider(root)
    s.pack(side=TOP)
    # Example 1: SupyForm
    labels=['Name', 'Age', 'Job']
    defaults=['Fridolin','40',"Großer Mandarin"]
    r=askSupyForm(labels,defaults,root,"Python ask form ...",'e',30,40)
    print(r)

    ##################################################################################################################################################################
    print("\nPart (iv): Test of SupySliderFrom")
    print("--------------------------------")
    ##################################################################################################################################################################
    # Example 2: SupySliderFrom
    labels=['var1','var2','var3','date']
    defaults=[3,14,5,datetime.date(2012,11,21)]
    ranges=[(0,10,1),(10,20,2),(5,10,0.1),(datetime.date(2012,1,1),datetime.date(2012,12,31))]
    r=askSupySliderForm(labels,defaults,ranges,root,"Slider Form example",'e',None,30,40)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import os.path
import sqlite3
from sqlite3 import OperationalError 
from supy.utilities import *

#########################################################################################################################
# supy.sqldatabase.py
#
# Simple database class for persistent storage of data records in files that is based on SQLITE
# 
# Database tables are defined with config classes (e.g., Table_config) that contain
#   - column names
#   - colum types
#   - column consraints (if any): note that any tables needs one primary key!
#   - column references (if any): foreign key reference, should be a only a table name
#   - column reference constraint (if any): foreign key constraints such as (on delete cascade)
#   - column default entries (if any)
#   - indexes
#
# 28.7.2014 - 30.7.2014 by Andreas Knoblauch
#########################################################################################################################


# ***************************************************************************
# class Table_cfg
# baseclass for database table specification
# ***************************************************************************
class Table_cfg:
    # table specification
    name = 'tb_demo'
    col_names           = ['id'         ,'name'      ,'address' ]
    col_types           = ['INTEGER'    ,'TEXT'      ,'TEXT'    ]
    col_references      = [None         ,None        ,None      ]
    col_constraints     = ['PRIMARY KEY','NOT NULL'  ,None      ]
    col_ref_constraints = [None         ,None        ,None      ]
    col_defaults        = [None         ,None        ,None      ]
    indexes             = [('idx_name',['name'])]
    # default rows (that will be stored during creation)
    rows_default     = ["(1,'Müller', 'Musterweg 5; 72475 Musterstadt')", \
                        "(2,'Maier' , 'Maierweg 6 ; 72473 Maierstadt' )"]
    # class methods
    @classmethod    
    def colnametupel(cls):
        return '(' + ','.join(n for n in cls.col_names) + ')'
    @classmethod    
    def colvaluetupel(cls,vallist):
        return '(' + ','.join(str(v) for v in vallist) + ')'


# ***************************************************************************
# class sqldatabase_cfg
# configuration of a complete sql database 
# - contains a (default) filename 
# - contains database specification by a list of table specifications 
#   (with derivatives of class Table_config)
# ***************************************************************************
class sqldatabase_cfg:
    default_filename = 'sqldatabase_demo.db'
    table_configs = [Table_cfg]


# ***************************************************************************
# class sqldatabase
# creation and management of a sql database 
# as specified by a derivative of sqldatabase_cfg  
# ***************************************************************************
class sqldatabase:
    def __init__(self,cfg=sqldatabase_cfg, filename=None, debugLevel=0):
        self.cfg=cfg
        self.filename=filename
        if(self.filename==None): self.filename=self.cfg.default_filename
        self.table_names = [tc.name for tc in self.cfg.table_configs]
        self.n_tables = len(self.table_names)
        self.cfg_of_table = {tc.name:tc for tc in self.cfg.table_configs}  # cfg_of_table['tablename'] is cfg of table with name 'tablename'
        self.idx_of_col   = {tc.name:{tc.col_names[i]:i for i in range(len(tc.col_names))} \
                                 for tc in self.cfg.table_configs }        # idx_of_column['tablename']['columnanme'] is index of column 'columnname' in table 'tablename'
        self.idx_of_pkey_cols = {tc.name:[i for i in range(len(tc.col_names)) if tc.col_constraints!=None and "PRIMARY KEY" in str(tc.col_constraints[i]).upper()]  \
                                 for tc in self.cfg.table_configs }        # idx_of_column['tablename']['columnanme'] is index of column 'columnname' in table 'tablename'
        self.db   = None            # reference to sql database
        self.curs = None            # reference to cursor to interact with sql database
        self.debugLevel=debugLevel
        self.connect()

    """
    connect to and initialize database
    """
    def connect(self):
        print("try to connect to fn=", self.filename)
        db1 = sqlite3.connect(self.filename)    # if database exists then just connect   
        db1.execute('pragma foreign_keys=ON')   # foreign key support necessary for DELETE CASCADE 
        self.db=db1                             # if no error occurs then assign db
        self.db.isolation_level = None          # necessary to have control over transactions
        self.curs=self.db.cursor()
        self.curs.execute("select name from sqlite_master WHERE type='table'")   # get all table names
        rowcount=len(self.curs.fetchall())
        if(rowcount!=self.n_tables):
            if(rowcount<=0):
                self.create_new_db()
            else:
                raise Exception("Database "+self.filename+" does not match specification!")

    """
    delete all tables of database
    """
    def delete_db(self):
        if(self.db):
            self.curs.execute("select name from sqlite_master WHERE type='table'")   # get all table names
            table_names = self.curs.fetchall()
            if len(table_names)>0:
                self.begin_transaction()
                for n in table_names:
                    cmd = "drop table " + n[0]
                    try:
                        if(self.debugLevel>0): print("EXECUTE SQL COMMAND: ", cmd)
                        self.curs.execute(cmd)
                    except OperationalError as e:
                        print("Error during execution of sqlite command: ", cmd)
                        raise e
                self.commit_transaction()

    """
    add/create new tables to database according to specifications in list tb_cfg
    """
    def add_tables(self,tb_cfgs):
        self.begin_transaction()
        for t in tb_cfgs: 
            # (i) create table t
            n=len(t.col_names)
            cmd = 'create table ' + t.name + ' ('
            for j in range(n):
                if(j>0): 
                    cmd=cmd+', '
                tb_name      = t.col_names[j]
                tb_type      = t.col_types[j]
                tb_ref       = t.col_references[j]
                tb_cnstr     = t.col_constraints[j]
                tb_ref_cnstr = t.col_ref_constraints[j]
                tb_def       = t.col_defaults[j]
                cmd = cmd + tb_name + ' ' + tb_type 
                if(tb_ref!=None)and(tb_ref!=''):
                    cmd = cmd + ' references ' + tb_ref
                    if(tb_ref_cnstr!=None)and(tb_ref_cnstr!=''):
                        cmd = cmd + ' ' + tb_ref_cnstr
                if(tb_cnstr!=None)and(tb_cnstr!=''):
                    cmd = cmd + ' ' + tb_cnstr 
                if(tb_def!=None):
                    cmd = cmd + " DEFAULT " + str(tb_def)
            cmd = cmd + ')'
            try:
                if(self.debugLevel>0): print("EXECUTE SQL COMMAND: ", cmd)
                self.curs.execute(cmd)
            except OperationalError as e:
                print("Error during execution of sqlite command: ", cmd)
                raise e
            else:
                # (ii) insert default rows in table t
                for r in t.rows_default:
                    cmd = 'insert into ' + t.name + ' ' + t.colnametupel() + ' values ' + r
                    try:
                        if(self.debugLevel>0): print("EXECUTE SQL COMMAND: ", cmd)
                        self.curs.execute(cmd)
                    except OperationalError as e:
                        print("Error during execution of sqlite command: ", cmd)
                        raise e
        # (iii) finalize transaction
        self.commit_transaction()
        self.resetIndexes()


    """
    create new (empty) database according to specification self.cfg.table_configs
    and insert default data records
    """
    def create_new_db(self):
        self.delete_db()    # delete all tables (if any)
        self.add_tables(self.cfg.table_configs)

    """
    create new (empty) database according to specification self.cfg.table_configs
    and insert default data records
    """
    def create_new_db_old(self):
        self.delete_db()    # delete all tables (if any)
        #print("begin_trans", len(self.cfg.table_configs))
        self.begin_transaction()
        for t in self.cfg.table_configs:
            # (i) create table t
            n=len(t.col_names)
            cmd = 'create table ' + t.name + ' ('
            for j in range(n):
                if(j>0): 
                    cmd=cmd+', '
                tb_name      = t.col_names[j]
                tb_type      = t.col_types[j]
                tb_ref       = t.col_references[j]
                tb_cnstr     = t.col_constraints[j]
                tb_ref_cnstr = t.col_ref_constraints[j]
                tb_def       = t.col_defaults[j]
                cmd = cmd + tb_name + ' ' + tb_type 
                if(tb_ref!=None)and(tb_ref!=''):
                    cmd = cmd + ' references ' + tb_ref
                    if(tb_ref_cnstr!=None)and(tb_ref_cnstr!=''):
                        cmd = cmd + ' ' + tb_ref_cnstr
                if(tb_cnstr!=None)and(tb_cnstr!=''):
                    cmd = cmd + ' ' + tb_cnstr 
                if(tb_def!=None):
                    cmd = cmd + " DEFAULT " + str(tb_def)
            cmd = cmd + ')'
            #print("cmd=",cmd)
            try:
                if(self.debugLevel>0): print("EXECUTE SQL COMMAND: ", cmd)
                self.curs.execute(cmd)
            except OperationalError as e:
                print("Error during execution of sqlite command: ", cmd)
                raise e
            # (ii) insert default rows in table t
            for r in t.rows_default:
                cmd = 'insert into ' + t.name + ' ' + t.colnametupel() + ' values ' + r
                try:
                    if(self.debugLevel>0): print("EXECUTE SQL COMMAND: ", cmd)
                    self.curs.execute(cmd)
                except OperationalError as e:
                    print("Error during execution of sqlite command: ", cmd)
                    raise e
        self.commit_transaction()
        self.resetIndexes()

    """
    drop all current indexes and reset new indexes as specified in self.cfg
    """
    def resetIndexes(self):
        self.begin_transaction()
        # (i) drop all indexes of the database
        self.curs.execute("select name from sqlite_master WHERE type='index'")   # get all index names
        index_names = self.curs.fetchall()
        for n in index_names:
            cmd = "drop index " + n[0]
            try:
                self.curs.execute(cmd)
            except OperationalError as e:
                print("Error during execution of sqlite command: ", cmd)
                raise e
        # (ii) create new indexes for all tables
        for t in self.cfg.table_configs:
            if t.indexes!=None:
                for i in range(len(t.indexes)):
                    n         = t.name+"_"+t.indexes[i][0]     # name of index
                    idx_cols  = t.indexes[i][1]     # columns of the index
                    if len(idx_cols)>0 : 
                        cmd = "create index " + n + " on " + t.name + "( "
                        for j in range(len(idx_cols)):
                            if j>0 : cmd = cmd + ", "
                            cmd = cmd + idx_cols[j] + " "
                        cmd = cmd + ")"
                        try:
                            self.curs.execute(cmd)
                        except OperationalError as e:
                            print("Error during execution of sqlite command: ", cmd)
                            raise e
        self.commit_transaction()
 
        
    """
    execute query and convert resulting list of row tuples to list of row dicts with field name keys
    thus, rowdicts[i][field] will return value of field of i-th record
    """
    def query2dicts(self,query,params=()):    
        try:
            self.curs.execute(query, params)
        except OperationalError as e:
            print("Error during execution of sqlite query '", query, "' with parameters: ", params)
            raise e
        colnames = [desc[0] for desc in self.curs.description]
        rowdicts = [dict(zip(colnames,row)) for row in self.curs.fetchall()]
        return rowdicts, colnames

    """
    formatted output of dict of records (e.g., from query2dictts)
    use optional parameter colnames to define an order of columns (otherwise alphabetical order)
    """
    def print_records(self,recs, colnames=None,sept=('-'*40)):
        print(len(recs), 'records')
        print(sept)
        #print("recs=", recs)
        for rec in recs:
            if(colnames):
                cn=colnames
            else:
                cn=rec
            maxkey = max(len(key) for key in cn)     # max key len
            for key in cn:
                print('%-*s => %s' % (maxkey, key, rec[key]))
            print(sept)

    """
    print table table_name
    """
    def print_table(self,table_name,format=True):
        curs=self.curs
        curs.execute('select count(*) from ' + table_name)
        nrec = curs.fetchone()[0]
        cmd = 'select * from ' + table_name
        curs.execute(cmd)
        colnames = [desc[0] for desc in self.curs.description]
        if not format:
            s1="Table %s (%d records)" % (table_name, nrec)
            s2="... with columns "+str(colnames)
            print(s1)
            print(s2)
            print(('-'*(max([len(s1)+1,len(s2)+1,15]))))
            while True:
                rec = curs.fetchone()
                if not rec: break
                print(rec)
        else:
            s1="Table %s" % (table_name)
            s2="... with columns "+str(colnames)
            print(s1)
            #print(s2)
            print(('-'*(max([len(s1)+1,len(s2)+1,15]))))
            print(('-'*(max([len(s1)+1,len(s2)+1,15]))))
            recs,cn = self.query2dicts(cmd) 
            self.print_records(recs,cn)

    """
    print database
    """
    def print_database(self,format=True):
        print("Database %s" % self.filename)
        print("... with Tables ", self.table_names)
        for t in self.table_names:
            self.print_table(t,format)
        
    """
    return string s as a list of elements from eltype (wrapper on supy.utilities method)
    useful to decode list attributes '[1,2,3,4]' which are stored as strings in the database tables,
    but should be transformed again as list of numerical elements after a query
    e.g., parseStringAsList('[1,2,3,4]','int') returns a integer list [1,2,3,4]
    eltype may be either 'int', 'float', 'string', 'binary' (or 'bool')
    in case of an excpetion use res_default as result of the transformation...
    """
    def parseStringAsList(self,s,eltype='int',res_default=None): 
        return parseStringAsList(s,eltype,res_default)

    """
    execute query and convert resulting list of row tuples to list of row dicts with field name keys
    thus, rowdicts[i][field] will return value of field of i-th record
    """
    def getColType(self,table,column):
        cfg_table=self.cfg_of_table[table]
        idx_col  =self.idx_of_col[table][column]
        return cfg_table.col_types[idx_col]

    """
    convert a column value to an appropriate string depending on column type,
    e.g., for type=INTEGER a value 5 becomes '5' (or '5' remains '5'), but for type=TEXT a value '5' must become '"5"'.
    """
    def toValueStr(self,table,col,val):
        coltype = self.getColType(table,col)
        v=str(val)               # per default just convert to str
        if(coltype=='TEXT'):
            v="'" + v + "'"
        return v

    """
    construct a simple where clause from matching table columns against column values
    Parameters:
       tables: list of tablenames
       cols  : list of column-names (for multiples entries cols[i] must be in the form 'table.colname')
       vals  : corresponding values (one data record from the external database
       oper  : operator for concatenating multiple conditions (default: AND)
    Returns: Where clause
    Example: wc=self.getWhereClause_from_data_matches(t,c,v) for c=['id','name'], v=[5,'Schmitz'] 
             returns wc=" id=5 AND name='Schmitz'"
    """
    def getWhereClause_from_ColumnValues(self,tables,cols, vals, oper="AND"):
        #print("getWhere... self=",self,"tables=",tables,"cols=",cols,"vals=",vals,"oper=",oper)
        str_error = 'AssertionError in sqldatabase.getWhereClause_from_ColumnValues(self,tables,cols, vals) for tables='+str(tables)+', cols='+str(cols)+', vals='+str(vals)+ " : "
        assert isinstance(tables,list) and isinstance(cols,list) and isinstance(vals,list), str_error + " All arguments must be lists!"
        assert len(cols)==len(vals), str_error + " List parameters cols and vals must have same length!"
        assert (len(cols)==0) or (len(tables)>0), str_error + " Missing table name!" 
        # (i) extract pure column names (from 'students.name' extract only 'name') 
        cols_only = [c for c in cols]     # copy of list
        tab_only  = [tables[0]]*len(cols)
        for i in range(len(cols)):
            if "." in cols[i]:
                splt=cols[i].split(".")
                cols_only[i]=splt[1]     # do not include table name
                tab_only [i]=splt[0]     # only table name
                assert tab_only[i] in tables, str_error + " table of " + str(cols[i]) + " is not in table list " + str(tables)
        # (ii) construct where clause
        wc=''
        for i in range(len(cols)):
            tab_cfg = self.cfg_of_table[tab_only[i]]
            assert cols_only[i] in tab_cfg.col_names, str_error + " Column " + str(cols_only[i]) + " does not exist in table " + str(tab_only[i]) 
            if wc!='': wc=wc+' '+oper+' '
            wc=wc+cols[i]+'=='
            coltype = self.getColType(tab_only[i],cols_only[i])
            #print("coltype=",coltype)
            if(coltype=='TEXT'):
                wc=wc+"'"+str(vals[i])+"'"
            else:
                wc=wc+str(vals[i])
        return wc


    """
    simple select default values of columns 
    parameters:
       colnames: list of column names of the columns to be selected
       tables: either a single table name or a list of tables for making a join
       strval: if True then all values are converted to strings
       strval: if True then all values are converted to strings
    """
    def simple_select_defaultvalues(self,colnames,tables,strval=False):
        str_error = 'AssertionError in sqldatabase.simple_select_defaultvalues(self,colnames,tables) for tables='+str(tables)+', colnames='+str(colnames)+ " : "
        assert isinstance(colnames,list), str_error+" Parameter colnames must be a list!"
        assert isinstance(tables,list) or isinstance(tables,str), str_error+" Parameter tables must be a string or a list!"
        res = [None for c in colnames]      # initialize results with list of None's
        for i in range(len(colnames)):
            # (a) extract table t and columname c
            c = colnames[i]
            if "." in c:
                splt=c.split(".")
                t=splt[0]
                c=splt[1]
                assert (t==tables) or (t in tables), str_error+" Table t=" + t + " does not match parameter tables!"
            else:
                t=tables
                if isinstance(tables,list): t=t[0]
            # (b) get default values
            cfg_table=self.cfg_of_table[t]
            idx_col  =self.idx_of_col[t][c]
            if(cfg_table.col_defaults!=None)and(cfg_table.col_defaults[idx_col]!=None):
                res[i]=cfg_table.col_defaults[idx_col]
                if(strval): res[i]=str(res[i])
                if(cfg_table.col_types[i]=='TEXT'): res[i]=res[i][1:-1]
        return res 
            

    """
    simple select query
    parameters:
       colnames: list of column names of the columns to be selected
       tables: either a single table name or a list of tables for making a join
       joinon: "join on" expression for each combination of tables (either None or a list being on item shorter than tables),
               e.g., tables=['t1','t2','t3'], join_on=['t1.a=t2.a', 't2.b=t3.b'] corresponds to 
                     SELECT .... FROM t1 join t2 on t1.a=t2.a join t3 on t2.b=t3.b ...
       where: where statement
       orderby: list of column names for sorting
       strval: if True then all values are converted to strings
    """
    def simple_select(self,colnames,tables,joinon=None,where=None,orderby=None,strval=False):
        # (i) decode column names
        if(colnames=='*'):
            str_colnames='*'
        else:
            str_colnames = ','.join(n for n in colnames)
        cmd = "select " + str_colnames + " from " 
        # (ii) decode table names and joins
        if isinstance(tables,str):
            str_tables = tables        # only single table (no join necessary)
        else:
            if joinon==None:
                # natural join
                str_tables = ' natural join '.join(n for n in tables)
            else:
                # general join
                for i in range(len(tables)):
                    if i==0:
                        str_tables=tables[0]
                    else:
                        str_tables = str_tables + " join " + tables[i] 
                        if len(joinon)>=i:
                            str_tables = str_tables + " on " + joinon[i-1]
        cmd = cmd + str_tables
        # (iv) decode where
        if(where!=None)and(where!=""):
            cmd = cmd + " where " + where
        # (iii) decode orderby
        if(orderby!=None):
            str_orderby = ','.join(n for n in orderby)
            cmd = cmd + " order by " + str_orderby 
        # (iv) execute command and convert to strings (if necessary)
        try:
            if(self.debugLevel>0): print("EXECUTE SQL COMMAND: ", cmd)
            self.curs.execute(cmd)
        except OperationalError as e:
            print("Error during execution of sqlite command: ", cmd)
            raise e
        data=self.curs.fetchall()
        if(self.debugLevel>1): print("RESULT OF SQL Command:\n", data)
        if (strval):  # convert all values to strings
            for i in range(len(data)):
                data[i]=list(data[i])
                data[i]=[str(d) for d in data[i]]
        return data


    """
    simple update query
    parameters:
       table: table to be updated
       colnames    : list of column names to be updated
       col_values  : list of column values corresponding to colnames
       pkeys       : list of columns acting as primary key to determine the row to be updated (in where clause)
       pkey_values : list of column values corresponding to pkeys
       orclause    : either 'rollback', 'abort', 'replace', 'fail', 'ignore', etc. 
    """
    def simple_update_byPKEY(self,table,colnames,col_values, pkeys, pkey_values, orclause=None):
        # (i) make assertions
        strerror="Assertion error in simple_update_byPKEY(self,table,colnames,col_values,pkeys,pkey_values) of table " \
            + str(table) + " with primary keys " + str(pkeys) + " = " + str(pkey_values) + ": "
        assert len(colnames)==len(col_values), strerror + "Parameters colnames and col_values must have same length!"
        assert len(pkeys)>0, strerror + "Parameter pkeys must contain at least one primary key!" 
        assert len(pkeys)==len(pkey_values), strerror + "Parameters pkeys and pkey_values must have same length!"
        # (ii) generate SQL query
        cmd = "update " 
        if orclause!=None : cmd = cmd + "or " + orclause + " "
        cmd = cmd + table + " set " 
        for i in range(len(colnames)):
            if(i>0): cmd = cmd + ", "
            cni=colnames[i]    # in case tab.col then remove tab.
            if "." in cni:
                splt=cni.split(".")
                cni=splt[1]    # only column name
            cmd = cmd + cni + " = " + self.toValueStr(table,cni,col_values[i])
        cmd = cmd + " where "
        for i in range(len(pkeys)):
            if(i>0): cmd = cmd + ", "
            cmd = cmd + pkeys[i] + " = " + self.toValueStr(table,pkeys[i],pkey_values[i])
        # (iii) perform update query
        try:
            if(self.debugLevel>0): print("EXECUTE SQL COMMAND: ", cmd)
            self.curs.execute(cmd)
        except OperationalError as e:
            print("Error during execution of sqlite command: ", cmd)
            raise e

    """
    simple insert query
    parameters:
       table: table where new record should be inserted into
       colnames    : list of column names where values are available 
       col_values  : list of column values corresponding to colnames
    """
    def simple_insert(self,table,colnames,col_values):
        # (i) make assertions
        strerror="Assertion error in simple_insert(self,table,colnames,col_values) of table " \
            + str(table) + " with colnames=" + str(colnames) + " and col_values=" + str(col_values) + ": "
        assert table in self.table_names, strerror + " Table " + str(table) + " is not database table!"
        assert (colnames==None) or (len(colnames)==len(col_values)), strerror + " Parameters colnames and col_values must have same length!"
        # (ii) generate SQL query
        cmd = "insert into " + table 
        if (colnames!=None) and (len(colnames)>0):
            cmd = cmd + " ("
            for i in range(len(colnames)):
                if(i>0): cmd=cmd + ", "
                cmd = cmd + colnames[i] 
            cmd = cmd + ")"
        cmd = cmd + " values ( "
        for i in range(len(col_values)):
            if(i>0): cmd=cmd + ", "
            cmd = cmd + self.toValueStr(table,colnames[i],col_values[i])
        cmd = cmd + " )"
        # (iii) perform insert query
        try:
            if(self.debugLevel>0): print("EXECUTE SQL COMMAND: ", cmd)
            self.curs.execute(cmd)
        except OperationalError as e:
            print("Error during execution of sqlite command: ", cmd)
            raise e

    """
    simple delete query
    parameters:
       table: table where record should be deleted
       colnames    : list of column names where values are available 
       col_values  : list of column values corresponding to colnames
    """
    def simple_delete(self,table,colnames,col_values):
        # (i) make assertions
        strerror="Assertion error in simple_delete(self,table,colnames,col_values) of table " \
            + str(table) + " with colnames=" + str(colnames) + " and col_values=" + str(col_values) + ": "
        assert table in self.table_names, strerror + " Table " + str(table) + " is not database table!"
        assert (colnames!=None) and (col_values!=None) and (len(colnames)==len(col_values)), strerror + " Parameters colnames and col_values must have same length!"
        # (ii) generate and perform SQL query
        cmd = "delete from " + table + " where " + self.getWhereClause_from_ColumnValues([table],colnames,col_values)
        try:
            self.curs.execute(cmd)
        except OperationalError as e:
            print("Error during execution of sqlite command: ", cmd)
            raise e

    """
    get_new_primary_key 
     - get new primary key for inserting a new data record into table
     - assumes that table has a unique primary key (i.e., only one column is primary key) 
     - here: assumes that the primary key has INTEGER-type !!
    """
    def get_new_primary_key(self,table, par_pkey_col=None):
        idx_of_pkey_cols = self.idx_of_pkey_cols[table]
        assert len(idx_of_pkey_cols)==1, "Currently only single-column primary-keys supported! However, table " + str(table) + " has multiple primary keys!"
        idx_of_pkey_col=idx_of_pkey_cols[0]           # only one primary key allowed!
        table_name = table                                          # name of the table
        pkey_col   = self.cfg_of_table[table].col_names[idx_of_pkey_col]  # column name of primary key
        assert par_pkey_col==None or par_pkey_col==pkey_col, "Column " + par_pkey_col + " of table " + table + " is not the primary key !!"
        cmd = 'select max(' + pkey_col + ') from ' + table
        try:
            if(self.debugLevel>0): print("EXECUTE SQL COMMAND: ", cmd)
            self.curs.execute(cmd)
        except OperationalError as e:
            print("Error during execution of sqlite command: ", cmd)
            raise e
        res=self.curs.fetchall()
        if(self.debugLevel>1):  print("RESULT OF SQL Command:\n", res)
        if(res==None) or (len(res)==0) or (res[0]==None) or (res[0][0]==None):
            res=0               # changed from res=1 to res=0 (19/9/2017)
        else:
            res=int(res[0][0])+1
        if(self.debugLevel>1):  print("REFINED RESULT OF SQL Command:\n", res)
        return res, pkey_col

    """
    begin_transaction 
    """
    def begin_transaction(self):
        self.curs.execute("begin transaction")

    """
    commit_transaction
    """
    def commit_transaction(self):
        self.curs.execute("commit transaction")

    """
    rollback_transaction
    """
    def rollback_transaction(self):
        self.curs.execute("rollback transaction")

    """
    set debug flag
    """
    def setDebugMode(self,debugLevel=0):
        self.debugLevel=debugLevel


#######################################
# Module test     
#######################################
if __name__ == '__main__':
    print("\nModule test of supylib module supy.sqldatabase.py")
    print("----------------------------------------------------\n") 
    db = sqldatabase()
    db.print_database(1)
    



#!/usr/bin/python
# -*- coding: utf-8 -*-
     
import os.path,sys,inspect, copy

def search_path(fname,paths=None):                   
    '''
    # if fname does not exist then try also paths in parameter paths and fname relative to module 
    '''
    # (i) ensure that path is a list
    if paths is None:
        paths=[]
    elif not isinstance(paths,list):
        paths=[paths]
    # (ii) append empty path (in order to search first to fname)
    paths = [""]+paths
    # (iii) append callers module paths to paths
    frame=inspect.currentframe()
    caller_paths=[]
    while not frame is None:
        code=frame.f_code
        co_filename = code.co_filename
        dr,fn = os.path.split(co_filename)
        caller_paths = [os.path.abspath(dr)]+caller_paths
        frame=frame.f_back
    paths = paths + caller_paths
    # (iv) now scan paths until file exists
    #print "scanning paths ", paths
    fname_result=None
    for p in paths:
        fname_result=os.path.join(p,fname)
        if os.path.exists(fname_result):
            break
    if(fname_result is None):
        print("supy.utilities.search_path: Cannot find file ", fname, " in paths ", paths, " !")
        fname_result=fname        # cannot find any better than fname
    return fname_result
        

def parseStringAsList(s,eltype='int',res_default=None): 
    """
    return string s as a list of elements from eltype useful to decode list attributes '[1,2,3,4]' which are stored as strings in the database tables,
    but should be transformed again as list of numerical elements after a query
    e.g., parseStringAsList('[1,2,3,4]','int') returns a integer list [1,2,3,4]
    eltype may be either 'int', 'float', 'string', 'binary' (or 'bool')
    in case of an excpetion use res_default as result of the transformation...
    """
    try:
        s=s.strip()    # remove white space from left and right
        s=s.lstrip('[')  # remove opening brackets from left
        s=s.rstrip(']')  # remove closing brackets form right
        sl = s.split(',')
        res=[]
        if eltype in ['int','integer','float','double','binary','bool','boolean']:            
            res=[float(ss.translate({ord(i):None for i in "'"})) for ss in sl]
            if eltype in ['int','integer','binary','bool','boolean']:
                res=[int(round(f)) for f in res]
        else:
            if eltype in ['string','text','str','txt']:
                res=[ss.strip() for ss in sl]           # remove white space (at beginning and ending of the substring)
                res=[ss.strip("'") for ss in sl]        # remove string delimiters (at beginning and ending of the substring)
    except ValueError as e:
        if(not res_default is None): 
            res=res_default
            print("Warning: ValueError in supy.utilities.parseStringAsList(...) for string s=",s,"eltype=",eltype,". Return default result res_default=",res_default)
        else:
            raise e
    return res


def asNumber(s,eltype='int',res_default=0): 
    """
    return string s as a single number of type eltype (int, float, bool, etc.)
    if error occurs return res_default
    e.g., asNumber('234','int')=234
          asNumber('234','float')=234.0
          asNumber('abcds','int',res_default=0)=0
    """
    n=res_default
    try:
        n=float(s)
        if eltype in ['int','integer','binary','bool','boolean']:
            n=int(round(n))
        if eltype in ['binary','bool','boolean']:
            n=bool(n)
    except ValueError as e:
        if(not res_default is None): 
            n=res_default
            print("Warning: ValueError in supy.utilities.asNumber(...) for string s=",s,"eltype=",eltype,". Return default result res_default=",res_default)
        else:
            raise e
    return n


def resource_path_pyinstaller(relative_path, default_path=None):  
    """
    transfers a relative path into an absolute path using PyInstallers _MEIPASS 
    """
    try:
        base_path = sys._MEIPASS    # PyInstaller creates a temp folder and stores path in _MEIPASS
    except Exception:
        if not default_path is None:
            base_path=default_path
        else:
            base_path = os.getcwd()     # previously: os.path.abspath(".")

    return os.path.join(base_path, relative_path)

def getAllFilenamesInFolder(path_to_folder,filetypes=['.png','.PNG','.jpg','.JPG','.tif','.TIF'],flagFullPaths=0):
    """
    get list of all filenames of desired types in the folder
    :param path_to_folder: path to the folder of the images
    :param filetypes: postfixes of the files to be selected
    :param flagFullPaths: If >0 then return list of full paths (pathtofolder+fname)
    :returns filelist: list of all filenames
    """
    filelist_all = os.listdir(path_to_folder)    # get list of all files in directory
    #print("filelist_all=\n",filelist_all)
    filelist = []
    for f in filelist_all:
        filename, fileextension = os.path.splitext(f)
        if fileextension in filetypes:
            if flagFullPaths>0:
                filelist=filelist+[os.path.join(path_to_folder,f)]
            else:
                filelist=filelist+[f]
    return filelist
    
if __name__ == '__main__':
    pass




#!/usr/bin/python
# -*- coding: utf-8 -*-

from Tkinter import *
from supy.database import *
from supy.multibase import *
from supy.forms import *
from supy.utilities import *
from copy import *
import tkMessageBox
import Image, ImageTk, ImageDraw

############################################################
# DBInputMask
# 28.3.2013 by Andreas Knoblauch
############################################################

class DBInputMask_config:
    # give options for each input column
    # attr_options[i] can be
    #   None       : no special options (plain input field)
    #   list       : make an OptionMenu to select from
    #   database   : make an OptionMenu from the keys of the database
    #   'filename' : string field with browse button
    #   'image'    : string field with browse button and image canvas
    #   'sound'    : string field with browse button (is played)
    #   any other string : is interpreted as name of a database in the multibase; make on OptionMenu from those keys
    attr_options = None
    attr_help_texts = None          # attr_helptext[i] can be used to give some help text
    attr_help_event = "<Button-3>"        # define event type to be bound to input label
    attr_help_callback = None       # define event handler (for None take default event handler)
    copy_default_flags = None       # if copy_default_flags[i]>0 then take for copy option the default attribute value
    image_size_field_idx = None     # input field index correponding to the image to be displayed (set None if not used) 
    imcanvas_size = [150,150]       # default size for the image canvas (if used)
    image_size_default = [100,100]  # image size if image_size_field_idx = None
    sound_playFlag = 0              # if set then sounds are played
    sound_primary_field_idx = None  # field index of primary sound that is played when entering new record (None if unused)
    recordlabel_txt = "DATA RECORD #"         # Label for Record headings
    recordlabel_font = ('courier',13,'bold')  # font for Record headings
    labelanchor='w'
    labelwidth=None                 # default; may be a list for each entry
    entrysize=40
    pos=None

    @classmethod
    def get_form_options(cls,mb=None):
        options=None # former: options=deepcopy(cls.attr_options) (failed sometimes, did not know why?)
        if cls.attr_options!=None:
            options=[None for i in range(len(cls.attr_options))]     
            for i in range(len(options)):
                if (cls.attr_options[i]=='image')or(cls.attr_options[i]=='sound')or(cls.attr_options[i]=='filename'):
                    options[i]='filename'
                else:
                    if isinstance(cls.attr_options[i],Database):
                        options[i]=cls.attr_options[i].get_list_of_keys()
                    elif isinstance(cls.attr_options[i],basestring):
                        options[i]=mb.get_database(cls.attr_options[i]).get_list_of_keys()
                    else:
                        options[i]=deepcopy(cls.attr_options[i])
        return options

    @classmethod
    def update_filename_entries(cls, db, old_base_path, new_base_path, enableCallbacks=0):
        '''
        updates all entries of filename attributes assuming
        that filenames in database are represented relative to old_base_path
        to new_base_path
        db is the database
        filename attributes are obtained from cfg.options 
        '''
        if cls.attr_options != None:
            idx_filename_attributes = [i for i, x in enumerate(cls.attr_options) if (x=='filename')or(x=='image')or(x=='sound')]
            if len(idx_filename_attributes)>0:
                oldflag=db.enableCallbacks(enableCallbacks)    # disable callbacks? Note that filenames do not really change, just relative paths!
                db_keys = db.get_list_of_keys()
                for k in db_keys:
                    for i in idx_filename_attributes:
                        rel_fname_old = db.get_record_entry(db.cfg.attr_names[i],k)
                        abs_fname = os.path.join(old_base_path,rel_fname_old)
                        rel_fname_new = os.path.relpath(abs_fname,new_base_path)
                        db.set_record_entry(db.cfg.attr_names[i],k,rel_fname_new)
                db.enableCallbacks(oldflag)      # enable callbacks again...

class DBInputMask:

    def __init__(self,parent,database,cfg=DBInputMask_config,multibase=None,paths_relative_to=os.getcwd()):
        self.parent = parent
        self.database = database
        self.cfg = cfg
        self.multibase = multibase   # set this if database is part of a multibase (required to get options form another db therein)
        self.paths_relative_to = paths_relative_to    # filenames are stored in database relative to this base path
        self.flagModified=0
        self.flag_singleton = self.database.cfg.flag_singleton  # only one database entry
        self.makeWidgets()

    def makeWidgets(self):
        # (i) root box container and frames for other widgets
        self.box = Frame(self.parent)
        self.box.pack(expand=YES, fill=X)
        upperframe = Frame(self.box)
        upperframe.pack(side=TOP,expand=Y,fill=X)
        middle1frame = Frame(self.box)
        middle1frame.pack(side=TOP,expand=Y,fill=X)
        middle2frame = Frame(self.box)
        middle2frame.pack(side=TOP,expand=Y,fill=X)
        lowerframe = Frame(self.box)
        lowerframe.pack(side=TOP,expand=Y,fill=X)
        # (ii) image canvas (displays only the first image field)
        self.imcanvas=None
        self.field_idx_image=0
        if(self.cfg.attr_options!=None)and('image' in self.cfg.attr_options):
            self.imcanvas = Canvas(middle2frame,width=self.cfg.imcanvas_size[0],height=self.cfg.imcanvas_size[1])
            self.imcanvas.pack(side=RIGHT,expand=Y,fill=X)
            for i in range(len(self.cfg.attr_options)):
                if(self.cfg.attr_options[i]=='image'):
                    self.field_idx_image = i
                    break;
        # (iii) Title label for records
        self.recordlabel = None
        if(self.cfg.recordlabel_txt!=""):
            self.recordlabel = Label(middle1frame,text="",font=self.cfg.recordlabel_font)
            self.recordlabel.pack(side=LEFT)
        # (iv) form 
        formframe = Frame(middle2frame)
        formframe.pack(side=LEFT,expand=Y,fill=X)
        self.form = SupyForm(self.database.cfg.attr_names,[""*self.database.attr_n],formframe,self.cfg.labelanchor,
                             labelwidth=self.cfg.labelwidth, entrysize=self.cfg.entrysize, pos=self.cfg.pos,
                             flagButtons=0,options=self.cfg.get_form_options(self.multibase),
                             paths_relative_to=self.paths_relative_to,browse_callback=self.browse_callback,
                             help_event=self.cfg.attr_help_event, help_texts=self.cfg.attr_help_texts, 
                             help_callback=self.cfg.attr_help_callback)
        if(self.flag_singleton):
            self.setFormEntries(self.database.get_record_entry())
        # (v) listbox for database keys
        if not self.flag_singleton:
            listframe = Frame(upperframe)
            listframe.pack(side=LEFT,expand=Y,fill=X)
            sbar = Scrollbar(listframe)
            self.listbox = Listbox(listframe,relief=SUNKEN,exportselection=0)
            list_of_keys = self.database.get_list_of_keys()
            for k in list_of_keys:
                self.listbox.insert('end',k)
            sbar.config(command=self.listbox.yview)
            self.listbox.config(yscrollcommand=sbar.set)
            sbar.pack(side=RIGHT, fill=Y)
            self.listbox.pack(side=LEFT, expand=YES, fill= BOTH)
            self.listbox.bind('<<ListboxSelect>>', self.onListboxSelect)
            if len(list_of_keys)>0 : 
                self.listbox.selection_set(first=0)
                self.onListboxSelect()
            self.listbox.focus_set()
        # (vi) create buttonframes next to listbox
        if not self.flag_singleton:
            buttonframe = Frame(upperframe)
            buttonframe.pack(side=RIGHT,expand=YES, fill=BOTH)
            self.var_radiobutton = IntVar()
            self.var_radiobutton.set(0)
            radiobuttonframe = Frame(buttonframe)
            radiobuttonframe.pack(side=TOP)
            Radiobutton(radiobuttonframe, text='view', command=self.onRadioButtonView,
                        variable=self.var_radiobutton,value=0).pack(side=LEFT)
            Radiobutton(radiobuttonframe, text='edit', command=self.onRadioButtonEdit,
                        variable=self.var_radiobutton,value=1).pack(side=LEFT)
            self.b_new    = Button(buttonframe,text='New'      ,command=self.onNew   )
            self.b_copy   = Button(buttonframe,text='Copy'     ,command=self.onCopy  )
            self.b_delete = Button(buttonframe,text='Delete'   ,command=self.onDelete)
            self.b_new.pack(side=TOP,fill=BOTH)
            self.b_copy.pack(side=TOP,fill=BOTH)
            self.b_delete.pack(side=TOP,fill=BOTH)
        # (vii) bottom buttonframe
        self.b_close=Button(lowerframe, text='Save changes & Close', command=self.onClose)
        self.b_close.pack(side=RIGHT)
        self.b_cancel=Button(lowerframe, text='Cancel changes & Close', command=self.onCancel)
        self.b_cancel.pack(side=RIGHT)
        if not self.flag_singleton:
            self.b_submit=Button(lowerframe, text='Submit Record', command=self.onSubmit)
            self.b_submit.pack(side=RIGHT)
        # (viii) disable input widgets (view only mode)
        if not self.flag_singleton:
            self.onRadioButtonView()

    def reset_listbox(self,idx_select):
        # delete entries of old listbox...
        #self.listbox.
        pass

    def setFormEntries(self,key):
        self.form.setEntryText(0,key)
        c = self.database.get_record_str(key)
        if(c!=None):           # no contents?
            for i in range(len(c)):
                self.form.setEntryText(i,c[i])
        else:
            self.form.setEntryText(0,key)
            for i in range(len(self.database.attr_n-1)):
                self.form.setEntryText(1+i,"")
        self.update_image()
        self.play_sound()

    def browse_callback(self,idx):
        if(idx==self.field_idx_image):
            self.update_image()
        if(self.cfg.attr_options!=None)and(self.cfg.attr_options[idx]=='sound'):
            self.play_sound(idx)

    def play_sound(self,idx=None):
        if(self.cfg.sound_playFlag>0):
            if(idx==None):
                idx=self.cfg.sound_primary_field_idx
            if(idx!=None):
                fname=self.form.entry_vars[idx].get()
                os.system('play ' + search_path(fname,self.paths_relative_to) + ' >/dev/null 2>/dev/null&')

    def load_image(self,fname,size):   # load image with filename fname and set to self.phim
        self.phim,self.draw=None,None
        if(fname):
            try:
                self.image = Image.open(search_path(fname,self.paths_relative_to))
            except IOError as e:
                print "I/O error({0}): {1}".format(e.errno, e.strerror)
                print "supy.database_inputmask.load_image: Cannot load image ", fname, " !"
                self.phim=None
            else:
                try:
                    self.image = self.image.resize(size)
                except TypeError as e:
                    print "I/O error({0}): {1}".format(e.errno, e.strerror)
                    print "Cannot display image ", fname, " !"
                    print "Invalid size parameter ", size
                    self.phim=None
                else:
                    self.draw = ImageDraw.Draw(self.image)
                    self.phim = ImageTk.PhotoImage(self.image)
        return self.phim,self.draw
        
    def update_image(self):
        if (self.imcanvas!=None):
            self.imcanvas.delete('all');
            fname=self.form.entry_vars[self.field_idx_image].get()
            try:
                sz_str=""
                if(self.cfg.image_size_field_idx!=None):
                    sz_str = self.form.entry_vars[self.cfg.image_size_field_idx].get()
                    sz=json.loads(sz_str)
                else:
                    sz=self.cfg.image_size_default
            except ValueError as e:
                print "Value error: "
                print "Maybe self.cfg.image_size_field_idx of Database ", self.cfg.fname, " points to wrong field ?"
                print "It should have a size content such as [50,50], "
                print "but instead has content sz_str=", sz_str
                print "Please check the configuration class of the Database!"
            else:
                self.load_image(fname,sz)     # sets self.phim
                posx,posy = int((self.cfg.imcanvas_size[0]-sz[0])/2),int((self.cfg.imcanvas_size[1]-sz[1])/2) 
                self.imcanvas.create_image(posx,posy,image=self.phim,anchor=NW)

    def onRadioButtonView(self):
        self.form.disableWidgets()
        if not self.flag_singleton:
            self.b_submit.configure(state=DISABLED)
            self.b_new   .configure(state=DISABLED)
            self.b_copy  .configure(state=DISABLED)
            self.b_delete.configure(state=DISABLED)

    def onRadioButtonEdit(self):
        self.form.enableWidgets()
        if not self.flag_singleton:
            self.b_submit.configure(state=NORMAL)
            self.b_new   .configure(state=NORMAL)
            self.b_copy  .configure(state=NORMAL)
            self.b_delete.configure(state=NORMAL)

    def onListboxSelect(self,evt=None):
        sels = self.listbox.curselection()
        if len(sels)>0:
            index = int(sels[0])
            key = self.listbox.get(index)
            if(self.recordlabel!=None):
                self.recordlabel.configure(text=self.cfg.recordlabel_txt+str(index+1))
            self.setFormEntries(key)

    def selectListboxItem(self,key):
        itemlist = self.listbox.get(0,END)
        if key in itemlist:
            idx = itemlist.index(key)
            self.listbox.selection_clear(first=0,last=END)
            self.listbox.selection_set(first=idx)
            self.listbox.activate(idx)
            self.listbox.see(idx)
            self.onListboxSelect()

    def onNew(self,call_onListboxSelect=1, call_onSubmit=1):
        new_key=self.database.add_default_record()
        self.listbox.insert(END,new_key)
        self.listbox.selection_clear(first=0,last=END)
        self.listbox.selection_set(first=END)
        self.listbox.activate(END)
        self.listbox.see(END)
        if(call_onListboxSelect>0): self.onListboxSelect()
        if(call_onSubmit>0)       : self.onSubmit()
        self.flagModified=1
    
    def onCopy(self):
        sel = self.listbox.curselection()
        if len(sel)>0:
            # get key of selected item
            index = int(sel[0])
            key   = self.listbox.get(index)
            content = self.database.get_record(key)[1:]
            # create new record
            self.onNew(0,0)
            # copy data 
            key_new = self.listbox.get(self.listbox.curselection()[0])
            content_default = deepcopy(self.database.get_record(key_new)[1:])  # content from default record   
            content_new     = deepcopy(content)                                # new content
            if self.cfg.copy_default_flags!=None:
                for i in range(len(self.cfg.copy_default_flags)):
                    if(i>0)and(self.cfg.copy_default_flags[i]>0):
                        content_new[i-1]=content_default[i-1]          # take default contents where flag is set
            self.database.set_record(key_new,*content_new)
            self.onListboxSelect()
            self.onSubmit()
            self.flagModified=1
            
    def onDelete(self):
        sel = self.listbox.curselection()
        if len(sel)>0:
            index = int(sel[0])
            key   = self.listbox.get(index)
            if self.database.delete_record(key):
                self.listbox.delete(index)
        else:
            index=0
        if index>=self.listbox.size():
            index=index-1
        self.listbox.selection_clear(first=0,last=END)
        if index>=0:
            self.listbox.selection_set(first=index)
            self.listbox.activate(index)
            self.onListboxSelect()
        self.flagModified=1
            
    def onSubmit(self):
        # get key of current data record
        key = None
        index = -1
        if self.flag_singleton:
            key = self.database.get_record_entry()    # selects first entry (i.e., key) of first (i.e., only) record
        else:
            sel = self.listbox.curselection()
            if len(sel)>0:
                # get key of selected item
                index = int(sel[0])
                key   = self.listbox.get(index)
        if key!=None:
            # create new database record
            key_new = self.form.entry_vars[0].get()
            content_new = ["" for i in range(self.database.attr_n-1)]
            for i in range(self.database.attr_n-1):
                content_new[i]=self.form.entry_vars[i+1].get()
            # check new database record (JSON may raise error)
            try:
                self.database.set_record_str([key_new]+content_new,1) # check only!!
            except ValueError as e:
                tkMessageBox.showerror("Submit error", "Cannot submit data. Please check your input!")
                print "Problem occured in database ", self.database.filename, " when submitting dataset ", [key_new]+content_new
            else:
                if not self.database.delete_record(key):
                    key_new=key                                  # old key must not be deleted; then use old key again...
                self.database.set_record_str([key_new]+content_new)
                self.flagModified=1
                # update listbox ?
                if not self.flag_singleton:
                    self.listbox.delete(index)
                    self.listbox.insert(index,key_new)
                    self.listbox.selection_set(index)
                    self.listbox.activate(index)
                    self.listbox.see(index)
                    self.onListboxSelect()
                    self.listbox.focus_set()

    def onClose(self):
        self.onSubmit()                 # submit latest changes in the form entry
        if(self.flagModified>0):
            self.database.save()
        self.box.destroy()

    def onCancel(self):
        if(self.flagModified>0):
            self.database.load()
        self.box.destroy()

def fillDBInputMask(parent,title,database,cfg=DBInputMask_config,multibase=None,paths_relative_to=os.getcwd()):
    top=Toplevel(parent)
    if parent: top.transient(parent)       # make window transient, e.g., minmize with parent etc.
    if title: top.title(title) 
    f=DBInputMask(top,database,cfg,multibase,paths_relative_to)
    top.wait_window(f.box)
    top.destroy()
    return None


if __name__ == '__main__':
    # create a dummy database
    class db_dummymask_config(Database_config,DBInputMask_config):
        # database specification
        default_filename = 'dummy_mask.db'
        dummy_img = '../MinotaurusSpiel/trunk/images/default_image.tif'
        dummy_snd = '../MinotaurusSpiel/trunk/sounds/default_sound.wav'
        types = ['type1','type2','type3']
        attr_names      = ['Nickname'  ,'Image'  , 'Imagesize', 'Typeinfo', 'Sound1' , 'Sound2' , 'Sound3' ]
        attr_types      = ['string'    ,'string' , 'json'     , 'string'  , 'string' , 'string' , 'string' ]
        default_record  = ['__new_key_',dummy_img, '[50,50]'  , types[0]  , dummy_snd, dummy_snd, dummy_snd]
        # input mask specification
        attr_options         = [None        ,'image'  , None       , types     , 'sound'  , 'sound'  , 'sound'  ]
        image_size_field_idx = 2
        sound_playFlag = 1
        sound_primary_field_idx = 4

    db_dummymask = Database(db_dummymask_config)
    db_dummymask.print_database()
    root = Tk()
    fillDBInputMask(root,"Edit database XY",db_dummymask,db_dummymask_config)
